// order numbers are from 1 !
#include <iostream>
#include <iomanip>
#include <fstream>
#include <string.h>
#include <stdlib.h>

#include "consts.h"
#include "defs.h"
#include "wprepea.h"
#include "wprepea_utils.h"
#include <set>
#include <map>



using namespace std;

map<wprepea, pair<wprepea, permutation> > isomap;

int lineno = 0;
bool parse_prepea(int num, ifstream &fin, signed char oplus[MAX][MAX], signed char ortholeft[MAX], signed char orthoright[MAX]) {

	char buffer[2048];
	int currenti = 0;

	fin.getline(buffer, 2048);
	lineno++;

	// skip empty lines and lines beginning with Order
	while (fin.good()) {
		// end of line
		if (!fin.good())
			return false;
		else {
			if (!strncmp(buffer, "Order", 5) || buffer[0]=='\0'
				|| !strncmp(buffer, "Pre", 3)) {
				fin.getline(buffer, 2048);
				lineno++;

				if (!fin.good()) 
					return false;
			}
			else
				break;

		}
	}

	if (!fin.good())
		return false;

	while (fin.good() && currenti < num) {
		char *pom = buffer;

#if 0
		if (!strncmp(pom,"Order", 5) || !*pom) {
			// skip line
			fin.getline(buffer, 2048);
			lineno++;
			continue;
		}
#endif
		for (int i = 0; i < num; i++) {
			// skip white spaces
			while (*pom == ' ' || *pom == '\t')
				pom++;
			
			int value = 0;
			char *pom_test = pom;
			
			if (*pom == '.') {
				oplus[currenti][i] = -1;
				pom++;
			}		
			else {
				while (*pom >= '0' && *pom <= '9') {
					value = 10*value + (*pom-'0');
					pom++;
				}
				if (pom_test == pom) {

					// no number was read
					cerr << " Error on the line " << lineno << " at position " << (int)(pom-buffer)+1 << "."<< endl << "Expecting number." << endl;
					cerr << buffer << endl;
					fin.close();
					return false;

				}
				else {
					oplus[currenti][i] = value;
				}
			}

			
		}
		fin.getline(buffer, 2048);
		lineno++;
		currenti++;
	}

	

		//next line Left orthosupplement
		char *pom = buffer;

		if (!strncmp(pom, "Left", 4)) {
			fin.getline(buffer, 2048);
			lineno++;
		}
		else {
			cerr << "Error: Expecting Left." << endl;
			fin.close();
			return false;
		}

		if (!fin.good()) {
			cerr << "Error: Expecting Left." << endl;
			fin.close();
			return false;
		}
		
		pom = buffer;

		for (int i = 0; i < num; i++) {
			// skip white spaces
			while (*pom == ' ' || *pom == '\t')
				pom++;
			
			int value = 0;
			char *pom_test = pom;
			while (*pom >= '0' && *pom <= '9') {
				value = 10*value + (*pom-'0');
				pom++;
			}
			if (pom_test == pom) {
				// no number was read
				cerr << " Error on the line " << lineno << " at position " << (int)(pom-buffer)+1 << "."<< endl << "Expecting number." << endl;
				fin.close();
				return false;
			}
			else {
				ortholeft[i] = value;
			}
		}
		fin.getline(buffer, 2048);
		lineno++;
		if (!fin.good()) {
			cerr << "Error: expecting Right." << endl;
			fin.close();
			return false;
		}

		if (!strncmp(buffer, "Right", 5)) {
			fin.getline(buffer, 2048);
			lineno++;
		}
		else {
			cerr << "Error: expecting Right." << endl;
			fin.close();
			return false;
		}
		if (!fin.good()) {
			cerr << "Error: expecting Right." << endl;
			fin.close();
			return false;
		}
		
		pom = buffer;

		for (int i = 0; i < num; i++) {
			// skip white spaces
			while (*pom == ' ' || *pom == '\t')
				pom++;
			
			int value = 0;
			char *pom_test = pom;
			while (*pom >= '0' && *pom <= '9') {
				value = 10*value + (*pom-'0');
				pom++;
			}
			if (pom_test == pom) {
				// no number was read
				cerr << " Error on the line " << lineno << " at position " << (int)(pom-buffer)+1 << "."<< endl << "Expecting number." << endl;
				fin.close();
				return false;
			}
			else {
				orthoright[i] = value;
			}
		}
	return true;
}


void load_initial_file(ifstream &fin, int num, int ord_number, set<wprepea> &setWPEs, set<wprepea> &setISOs) {

	signed char oplus[MAX][MAX];
	signed char ortholeft[MAX];
	signed char orthoright[MAX];
	while (parse_prepea(num, fin, oplus, ortholeft, orthoright)) {

		for (int i = 0 ; i < num ; i++) {
			for (int j = 0 ; j < num ; j++) 
				if (oplus[i][j] < 0) 
					cout << setw(3) << ".";
				else 
					cout << setw(3) << (signed int)oplus[i][j] ; 
			cout << endl;
		}

		cout << "Left orthosupplement:" << endl;
		for (int i = 0 ; i < num ; i++) 
			cout << setw(3) << (signed int)ortholeft[i] ;
		cout << endl;

		cout << "Right orthosupplement:" << endl;
		for (int i = 0 ; i < num ; i++) 
			cout << setw(3) << (signed int)orthoright[i] ;
		cout << endl;

		if (condition(num, oplus)) {
			cout << "Pre pseudo effect algebra." << endl;
		}

		cout << endl;

		wprepea tmpWP(num, oplus, ortholeft, orthoright);

		if (setWPEs.find(tmpWP) == setWPEs.end() && setISOs.find(tmpWP) == setISOs.end()) {
			setWPEs.insert(tmpWP);

			//update_ISOS(ORDERSARRAY[ord_number], tmpWP, setWPEs, setISOs);

			update_ISOS_with_map(ORDERSARRAY[ord_number], tmpWP, setWPEs, setISOs, isomap);

		}
		else {
			cerr << "Warning: Loaded file contains either duplicates or isomorphic copies of certain Weak Pre Pseudo Effect Algebras models" << endl;
		}
	}
	
}

void merge_file(string filename, int num, int ord_number, set<wprepea> &setWPEs, set<wprepea> &setISOs) {

	ifstream fin;
	fin.open(filename.c_str(), ifstream::in);
	
	if (!fin.good()) {
		cerr << "Error opening file `" << filename << "`." << endl;
		return;
	}

	cerr << "Merging file " << filename << "." << endl;
	signed char oplus[MAX][MAX];
	signed char ortholeft[MAX];
	signed char orthoright[MAX];
	while (parse_prepea(num, fin, oplus, ortholeft, orthoright)) {


		wprepea tmpWP(num, oplus, ortholeft, orthoright);
		set<wprepea>::const_iterator findWPEs = setWPEs.find(tmpWP);
		set<wprepea>::const_iterator findISOs = setISOs.find(tmpWP);

		if ( findWPEs == setWPEs.end() && findISOs == setISOs.end()) {
			for (int i = 0 ; i < num ; i++) {
				for (int j = 0 ; j < num ; j++) 
					if (oplus[i][j] < 0) 
						cout << setw(3) << ".";
					else 
						cout << setw(3) << (signed int)oplus[i][j] ; 

				cout << endl;
			}

			cout << "Left orthosupplement:" << endl;
			for (int i = 0 ; i < num ; i++) 
				cout << setw(3) << (signed int)ortholeft[i] ;
			cout << endl;

			cout << "Right orthosupplement:" << endl;
			for (int i = 0 ; i < num ; i++) 
				cout << setw(3) << (signed int)orthoright[i] ;
			cout << endl;

			if (condition(num, oplus)) {
				cout << "Pre pseudo effect algebra." << endl;
			}
			cout << endl;

			setWPEs.insert(tmpWP);


			update_ISOS_with_map(ORDERSARRAY[ord_number], tmpWP, setWPEs, setISOs, isomap);


//			update_ISOS(ORDERSARRAY[ord_number], tmpWP, setWPEs, setISOs);
		}
		else {
			
			cerr << "Isomorphic copy already in set: " << endl << endl;
			print_op(cerr, tmpWP.n, tmpWP.oplus);
			cerr << "Left orthosupplement:" << endl;
			for (int i = 0 ; i < tmpWP.n ; i++) 
				cerr << setw(3) << (signed int)tmpWP.left_ortho[i] ;
			cerr << endl;

			cerr << "Right orthosupplement:" << endl;
			for (int i = 0 ; i < tmpWP.n ; i++) 
				cerr << setw(3) << (signed int)tmpWP.right_ortho[i] ;
			cerr << endl;


			map<wprepea, pair<wprepea, permutation> >::const_iterator it;

			it = isomap.find(tmpWP);
			if (it != isomap.end()) {
				cerr << endl << " is isomorphic to " << endl << endl;

				print_op(cerr, ((it->second).first).n, ((it->second).first).oplus);

				cerr << "Left orthosupplement:" << endl;
				for (int i = 0 ; i < ((it->second).first).n ; i++) 
					cerr << setw(3) << (signed int)((it->second).first).left_ortho[i] ;
				cerr << endl;

				cerr << "Right orthosupplement:" << endl;
				for (int i = 0 ; i < ((it->second).first).n ; i++) 
					cerr << setw(3) << (signed int)((it->second).first).right_ortho[i] ;
				cerr << endl;

				cerr << "\tvia isomorphism"<< endl;
				for (int i=0; i < ((it->second).second).length; i++)
					cerr << setw(3) << ((it->second).second).perm[i] ;
				cerr << endl;
				cerr << "---" << endl;
			}
			else {
				cerr << endl << "is already in set." << endl;
				cerr << "---" << endl;
			}
		
		}
	}
	
	fin.close();
}

int main(int argc, char** argv) {
	ifstream fin;
	set<struct wprepea> setWPEs, setISOs;


	if (argc < 3) {
		cout << "Usage: " << argv[0] << " <order> <file_name> [<merge_file_1> ... <merge_file_n>]" << endl;
		cout << "Orders range: " << VFROM + 1 << " - " << VTO << endl;
	}
	else {
		int num = N;
		int num_from = atoi(argv[1]) - 1;
		if (num_from < VFROM) {
			cout << "Orders range: " << VFROM + 1 << " - " << VTO << endl;
			return 1;
		}

		if (num_from >= VTO) {
			cout << "Orders range: " << VFROM + 1 << " - " << VTO << endl;
			return 1;
		}

		int order_number = num_from - order_offset;
		int idx = 3;

		fin.open(argv[2], ifstream::in);
		if (fin.good()) {

			cout << "Order n. " << num_from + 1 << endl;

			cerr << "Loading initial file '" << argv[2] << "'." << endl;
			load_initial_file(fin, num, order_number, setWPEs, setISOs); 

			fin.close();

			while (idx < argc) {
				merge_file(argv[idx], num, order_number, setWPEs, setISOs);
				idx++;
			}

		}
		else {
			cerr << "Error: Cannot open '" << argv[2] << "' file." << endl;
			return 1;
		}
	}
	return 0;
}
