// order numbers are from 1 !
#include <iostream>
#include <iomanip>
#include <fstream>
#include <string.h>
#include <stdlib.h>

#include "consts.h"
#include "defs.h"
#include "wprepea.h"
#include "wprepea_utils.h"
#include <set>
#include <map>



using namespace std;

int lineno = 0;

void print_op(ostream& o, signed char n, const signed char op[MAX][MAX]) {
	for (int i = 0; i < n; i++) {
		for (int j = 0; j < n; j++) {

			if (is_defined(op, i, j))
				o << setw(3) << (int)get_oplus(op, i, j);
			else
				o << setw(3) << ".";
		}
		o << endl;
	}
}


bool is_commutative(int n, signed char op_plus[MAX][MAX]) {
	for (int i = 1; i < n - 1; i++)
	for (int j = 1; j < n - 1; j++)
		if (is_defined(op_plus, i, j) && (!is_defined(op_plus, j, i) || get_oplus(op_plus, i, j) != get_oplus(op_plus, j, i)))
			return false;

	return true;
}

bool condition(int n, signed char op_plus[MAX][MAX]) {
	for (int i = 0; i < n; i++)
	for (int j = 0; j < n; j++)
		if (is_defined(op_plus, i, j)) {
			bool found_c = false;
			int result = get_oplus(op_plus, i, j);

			for (int k = 0; k < n; k++)
				if (get_oplus(op_plus, k, i) == result) {
					found_c = true;
					break;
				}
			if (!found_c)
				return false;

			bool found_d = false;


			for (int k = 0; k < n; k++)
				if (get_oplus(op_plus, j, k) == result) {
					found_d = true;
					break;
				}
			if (!found_d)
				return false;
		}
	return true;
}

bool parse_prepea(int num, ifstream &fin, signed char oplus[MAX][MAX], signed char ortholeft[MAX], signed char orthoright[MAX]) {

	char buffer[2048];
	int currenti = 0;

	fin.getline(buffer, 2048);
	lineno++;

	// skip empty lines and lines beginning with Order
	while (fin.good()) {
		// end of line
		if (!fin.good())
			return false;
		else {
			if (!strncmp(buffer, "Order", 5) || buffer[0]=='\0'
				|| !strncmp(buffer, "Pre", 3)) {
				if (!strncmp(buffer, "Order", 5))
					cout << buffer << endl;

				fin.getline(buffer, 2048);
				lineno++;

				if (!fin.good()) 
					return false;
			}
			else
				break;

		}
	}

	if (!fin.good())
		return false;

	while (fin.good() && currenti < num) {
		char *pom = buffer;

#if 0
		if (!strncmp(pom,"Order", 5) || !*pom) {
			// skip line
			fin.getline(buffer, 2048);
			lineno++;
			continue;
		}
#endif
		for (int i = 0; i < num; i++) {
			// skip white spaces
			while (*pom == ' ' || *pom == '\t')
				pom++;
			
			int value = 0;
			char *pom_test = pom;
			
			if (*pom == '.') {
				oplus[currenti][i] = -1;
				pom++;
			}		
			else {
				while (*pom >= '0' && *pom <= '9') {
					value = 10*value + (*pom-'0');
					pom++;
				}
				if (pom_test == pom) {

					// no number was read
					cerr << " Error on the line " << lineno << " at position " << (int)(pom-buffer)+1 << "."<< endl << "Expecting number." << endl;
					cerr << buffer << endl;
					fin.close();
					return false;

				}
				else {
					oplus[currenti][i] = value;
				}
			}

			
		}
		fin.getline(buffer, 2048);
		lineno++;
		currenti++;
	}

	

		//next line Left orthosupplement
		char *pom = buffer;

		if (!strncmp(pom, "Left", 4)) {
			fin.getline(buffer, 2048);
			lineno++;
		}
		else {
			cerr << "Error: Expecting Left." << endl;
			fin.close();
			return false;
		}

		if (!fin.good()) {
			cerr << "Error: Expecting Left." << endl;
			fin.close();
			return false;
		}
		
		pom = buffer;

		for (int i = 0; i < num; i++) {
			// skip white spaces
			while (*pom == ' ' || *pom == '\t')
				pom++;
			
			int value = 0;
			char *pom_test = pom;
			while (*pom >= '0' && *pom <= '9') {
				value = 10*value + (*pom-'0');
				pom++;
			}
			if (pom_test == pom) {
				// no number was read
				cerr << " Error on the line " << lineno << " at position " << (int)(pom-buffer)+1 << "."<< endl << "Expecting number." << endl;
				fin.close();
				return false;
			}
			else {
				ortholeft[i] = value;
			}
		}
		fin.getline(buffer, 2048);
		lineno++;
		if (!fin.good()) {
			cerr << "Error: expecting Right." << endl;
			fin.close();
			return false;
		}

		if (!strncmp(buffer, "Right", 5)) {
			fin.getline(buffer, 2048);
			lineno++;
		}
		else {
			cerr << "Error: expecting Right." << endl;
			fin.close();
			return false;
		}
		if (!fin.good()) {
			cerr << "Error: expecting Right." << endl;
			fin.close();
			return false;
		}
		
		pom = buffer;

		for (int i = 0; i < num; i++) {
			// skip white spaces
			while (*pom == ' ' || *pom == '\t')
				pom++;
			
			int value = 0;
			char *pom_test = pom;
			while (*pom >= '0' && *pom <= '9') {
				value = 10*value + (*pom-'0');
				pom++;
			}
			if (pom_test == pom) {
				// no number was read
				cerr << " Error on the line " << lineno << " at position " << (int)(pom-buffer)+1 << "."<< endl << "Expecting number." << endl;
				fin.close();
				return false;
			}
			else {
				orthoright[i] = value;
			}
		}
	return true;
}


bool is_pea(int num, signed char op_plus[MAX][MAX]) {

	for (int i = 0; i < num; i++) {	
		int count = 0;

		for (int j = 0; j < num; j++) {
			if (get_oplus(op_plus, i, j) == num - 1)
				count++;
		}

		if (count > 1) 
			return false;
		
	}
	return true;
}

bool custom_test(int num, signed char op_plus[MAX][MAX], signed char ortholeft[MAX], signed char orthoright[MAX]) {
// only 1 line after return should be uncommented
	return 
	// condition(num, op_plus); // pre pseudo effect algebra

	// is_commutative(num, op_plus); // commutative weak pre pseudo effect algebra = 
	// is_commutative(num, op_plus) && condition(num, op_plus); // pre effect algebra  

	// is_pea(num, op_plus) && condition(num, op_plus); // pseudo effect algebra
	 is_commutative(num, op_plus) && is_pea(num, op_plus) && condition(num, op_plus); // effect algebra
	//true; // weak pre pseudo effect algebras
}


void load_file_and_test(ifstream &fin, int num) {

	signed char oplus[MAX][MAX];
	signed char ortholeft[MAX];
	signed char orthoright[MAX];

	while (parse_prepea(num, fin, oplus, ortholeft, orthoright)) {

		// Perform Tests

		if (custom_test(num, oplus, ortholeft, orthoright)) {

			for (int i = 0 ; i < num ; i++) {
				for (int j = 0 ; j < num ; j++) 
					if (oplus[i][j] < 0) 
						cout << setw(3) << ".";
					else 
						cout << setw(3) << (signed int)oplus[i][j] ; 
				cout << endl;
			}

			cout << "Left orthosupplement:" << endl;
			for (int i = 0 ; i < num ; i++) 
				cout << setw(3) << (signed int)ortholeft[i] ;
			cout << endl;

			cout << "Right orthosupplement:" << endl;
			for (int i = 0 ; i < num ; i++) 
				cout << setw(3) << (signed int)orthoright[i] ;
			cout << endl;

	//		if (condition(num, oplus)) {
	//			cout << "Pre pseudo effect algebra." << endl;
	//		}
			cout << endl;
		}
	}
	
}

int main(int argc, char** argv) {
	ifstream fin;
	set<struct wprepea> setWPEs, setISOs;


	if (argc < 3) {
		cout << "Usage: " << argv[0] << " <num> <file_name>" << endl;
		cout << "<n> = 3-11." << endl;
		cout << "Loads all weak pre pseudo effect algebras from" << endl <<  
			"file <file_name> and test their properties." <<endl;
	}
	else {
		int num = atoi(argv[1]);
		fin.open(argv[2], ifstream::in);
		if (fin.good()) {

			load_file_and_test(fin, num); 

			fin.close();
		}
		else {
			cerr << "Error: Cannot open '" << argv[2] << "' file." << endl;
			return 1;
		}
	}
	return 0;
}
