#include <time.h>
#include <stdlib.h>
#include <iostream>
#include <iomanip>
#include <set>

#include "consts.h"
#include "wprepea.h"
#include "wprepea_utils.h"
#include "defs.h"


#define ITERCOUNT 100000000UL
#undef DEBUG

using namespace std;

// global
unsigned long int no_of_iterations = 1;

void generate_all_orthosupplements(signed char n, signed char op_plus[MAX][MAX], signed char left_ortho[MAX], signed char right_ortho[MAX]);
//bool is_defined(const int op_plus[MAX][MAX], int elem_a, int elem_b);
void dump_state(signed char n, signed char op_plus[MAX][MAX]); 




//bool operator<(const struct wprepea &wp1, const struct wprepea &wp2); 

struct wprepeaCompare {
	bool operator() (const struct wprepea &wp1, const struct wprepea &wp2) const;
};

/*bool wprepeaCompare::operator() (const struct wprepea &wp1, const struct wprepea &wp2) const {
	return operator<(wp1, wp2);
}
*/

void dump_state(signed char n, signed char op_plus[MAX][MAX]) {
	cerr << "OP: ";
	for (int i = 1; i < n-1; i++)
		for (int j = 1; j < n-1;j++)
			cerr << (int)op_plus[j][i] << ", " ;
	cerr << endl;
}

bool test_partial_order_right(signed char n, signed char op_plus[MAX][MAX], signed char right_ortho[MAX]) {


	// if a + b~ and b + a~ are defined and a != b
	// antisymmetry
	for (int elem_a = 1; elem_a < n - 1; elem_a++)
	for (int elem_b = 1; elem_b < n - 1; elem_b++)
		if (is_defined(op_plus, elem_a, right_ortho[elem_b]) &&
			is_defined(op_plus, elem_b, right_ortho[elem_a]) &&
			elem_a != elem_b
			)
			return false;

	// transitivity
	// a R b and b R c => a R c
	for (int elem_a = 1; elem_a < n - 1; elem_a++)
	for (int elem_b = 1; elem_b < n - 1; elem_b++)
	for (int elem_c = 1; elem_c < n - 1; elem_c++)
		if ( is_defined(op_plus, elem_a, right_ortho[elem_b]) &&
		     is_defined(op_plus, elem_b, right_ortho[elem_c]) &&
		    !is_defined(op_plus, elem_a, right_ortho[elem_c])
		)
			return false;

	return true;
}

bool test_partial_order_left(signed char n, signed char op_plus[MAX][MAX], signed char left_ortho[MAX]) {

	// antisymmetry
	for (int elem_a = 1; elem_a < n - 1; elem_a++)
	for (int elem_b = 1; elem_b < n - 1; elem_b++)
		if (is_defined(op_plus, left_ortho[elem_b], elem_a) &&
			is_defined(op_plus, left_ortho[elem_a], elem_b) &&
			elem_a != elem_b
			) {
			cout << "Left order Antisymmetry failed for: " << elem_a << ", " << elem_b << "." << endl;
			return false;
		}

	// transitivity
	for (int elem_a = 1; elem_a < n - 1; elem_a++)
	for (int elem_b = 1; elem_b < n - 1; elem_b++)
	for (int elem_c = 1; elem_c < n - 1; elem_c++)
		if ( is_defined(op_plus, left_ortho[elem_b], elem_a) &&
			 is_defined(op_plus, left_ortho[elem_c], elem_b) &&
			!is_defined(op_plus, left_ortho[elem_c], elem_a)
		) {
			cout << "Left order Transitivity failed for: " << elem_a << ", " << elem_b << ", " << elem_c << "." << endl;
			return false;
		}

#ifdef DEBUG
	cout << "Left order Reflexivity: OK." << endl;
	cout << "Left order Antisymmetry: OK." << endl;
	cout << "Left order Transitivity: OK." << endl;
#endif
	return true;
}


bool exists_left_orthosupplement(signed char n, signed char op_plus[MAX][MAX]) {
	for (int i = 0; i < n; i++) {
		bool found = false;

		for (int j = 0; j < n; j++)
			if (get_oplus(op_plus, j, i) == n - 1) {
				found = true;
				break;
			}
		if (!found)
			return false;
	}
	return true;
}

bool exists_right_orthosupplement(signed char n, signed char op_plus[MAX][MAX]) {
	for (int i = 0; i < n; i++) {
		bool found = false;

		for (int j = 0; j < n; j++)
			if (get_oplus(op_plus, i, j) == n - 1) {
				found = true;
				break;
			}
		if (!found)
			return false;
	}
	return true;
}

void print_set_todo(set<todoElem> &set_todo) {
	cerr << "n: " << set_todo.size() << endl;
	for (set<todoElem>::iterator it = set_todo.begin(); it != set_todo.end(); it++) 
		cerr << "(" << (int)it->x << "," << (int)it->y << "), ";
	
	cerr << endl;
}

void gen_all_permutations_set_plan(int level, signed char n, signed char perm[MAX], int ordersmall[MAX][MAX], set<struct wprepea> &setWPES, set<struct wprepea> &setISOS, set<setpermElem> &setPERM) {
	int tmp;

	if (level == n-1) {
		signed char right_ortho[MAX];

		compute_inverse_perm(n, perm, right_ortho);

		set<todoElem> test2;


		if (test_order_ortho(n, ordersmall, perm, right_ortho)) {


			if (
			 init_set_todo(n, test2, ordersmall, perm, right_ortho) &&
				setPERM.find(setpermElem(n,perm)) == setPERM.end()
			) {

				cerr << "Tested permutation: ";
				for (int i = 1; i < n-1;i++)
					cerr << setw(3) << (int)perm[i];
				cerr << endl;

				print_set_todo(test2);

				signed char perm_iso[MAX];

				for (int i = 0; i < n; i++) 
					perm_iso[i] = i;

				update_set_perm(1, ordersmall, setPERM, n, perm, perm_iso, true);
				}
			}

	}
	else {
		for (int idx=level; idx < n-1; idx++) {
			if (idx != level) {
				tmp = perm[idx];
				perm[idx] = perm[level];
				perm[level] = tmp;

			}
			gen_all_permutations_set_plan(level+1, n, perm, ordersmall, setWPES, setISOS, setPERM);

			if (idx != level) {
				tmp = perm[idx];
				perm[idx] = perm[level];
				perm[level] = tmp;
			}
		}
	}	


}






int main(int argc, char* argv[]) {

	int n = N;

	signed char perm[MAX];
	perm[0] = n - 1; perm[n - 1] = 0;


	for (int idx = (VFROM-order_offset); idx < (VTO-order_offset); idx++) {
		set<struct wprepea> setWPES, setISOS;
		set<setpermElem> setPERM;
		
		for (int i = 1; i < n - 1; i++) perm[i] = i;

		cout << "Order n. " << order_offset + idx + 1 << endl;
		cerr << "Order n. " << order_offset + idx + 1 << endl;


		for (int a = 0; a < n - 2; a++) {
			for (int b = 0; b < n - 2; b++)
				cerr << setw(2) << ORDERSARRAY[idx][a][b];
			cerr << endl;
		}


		gen_all_permutations_set_plan(1, n, perm, ORDERSARRAY[idx], setWPES, setISOS, setPERM);
	}

	return 0;
}
