#include <time.h>
#include <stdlib.h>
#include <iostream>
#include <iomanip>
#include <set>
#include <map>

#include "consts.h"
#include "defs.h"
#include "wprepea.h"
#include "wprepea_utils.h"
#include "wprepea_parseutils.h"

#define ITERCOUNT 100000000UL
#undef DEBUG

using namespace std;

// global
unsigned long int no_of_iterations = 0;

void gen_oplus_from_set_todo_from_state_to_state_with_depMap_with_assocCheckMap(set<todoElem>::iterator it, signed char n, set<todoElem> &set_todo, int ordersmall[MAX][MAX], signed char oplus[MAX][MAX], signed char perm_left[MAX], signed char perm_right[MAX], set<struct wprepea> &setWPES, set<struct wprepea> &setISOS, int cur_depth, int *from_state, bool *from_was_used, int *to_state, bool *to_was_used, const map<todoElem, set<todoElem> > &dependencyMap, int *cur_state, map< pair<signed char, signed char>, set<signed char> > &assocCheckMap) {

	if (it == set_todo.end()) {
#ifdef TESTOUT
	cerr << "\t\tTesting the state: ";
	set<todoElem>::iterator itertmp = set_todo.begin();
	while (itertmp != set_todo.end()) {
		cerr << (int)oplus[itertmp->x][itertmp->y] << ", ";
		++itertmp;
	}
	cerr << endl;
#endif
		if (no_of_iterations == ITERCOUNT) {
			time_t tim = time(NULL);
			cerr << "[" << tim << "]" << endl;
			no_of_iterations = 0;
			cerr << "\tState: ";
			for (set<todoElem>::iterator itp = set_todo.begin(); itp != set_todo.end(); ++itp)
				cerr << setw(2) << (int)oplus[itp->x][itp->y];
			cerr << endl;	
		}
		no_of_iterations++;

		if (is_weak_pre_pea(n, oplus, perm_right, perm_left)) {

			struct wprepea wp(n, oplus, perm_left, perm_right);
			set<struct wprepea>::iterator it1, it2;

			it1 = setWPES.find(wp);

			it2 = setISOS.find(wp);

			if (it1 == setWPES.end() && it2 == setISOS.end()) {

				setWPES.insert(wp);

				print_op(cout, n, oplus);

				cout << "Left orthosupplement:"<<endl;
				for (int i = 0; i < n; i++)
					cout << setw(3) << (int)perm_left[i];
				cout << endl;

				cout << "Right orthosupplement:"<<endl;
				for (int i = 0; i < n; i++)
					cout << setw(3) << (int)perm_right[i];
				cout << endl << endl;
				
#ifdef DEBUG
				cout << "Storing into WPES set." << endl;
#endif

				
				if (condition(n, oplus))
					cout << "Pre pseudo effect algebra." << endl;
				cout << endl;	
				
				update_ISOS(ordersmall, wp, setWPES, setISOS);
			}


		}
	}	
	else {

		signed char atleast = max(it->x, it->y), 
		    atmost = n - 1, start_from, end_to;

		if (!from_was_used[cur_depth]) {
			start_from = from_state[cur_depth];
			from_was_used[cur_depth] = true;
		}
		else {
			start_from = atleast;
		}

		if (!to_was_used[cur_depth]) {

			bool is_to_state_prefix = true;
			for (int i = 0; i < cur_depth; i++)
				if (cur_state[i] != to_state[i]) {
					is_to_state_prefix = false;
					break;

				}

			if (is_to_state_prefix) {
				end_to = to_state[cur_depth];
				to_was_used[cur_depth] = true;
			}
			else 
				end_to = atmost;
		}
		else 
			end_to = atmost;

		#ifdef DEBUG
		//		cout << "Level: " << level << endl << "At least: " << atleast << endl << "At most: " << atmost << endl;
		#endif

		for (int idx = start_from ; idx <= end_to; idx++) {
			if (idx != it->x && idx != it->y 
 				//(atleast == n - 1 || idx == n - 1 || ordersmall[atleast - 1][idx - 1]) 
				&& (idx == n - 1 || ordersmall[it->x - 1][idx - 1]) 
				&& (idx == n - 1 || ordersmall[it->y - 1][idx - 1])
				&& (idx == n - 1 || isFromUpperBound(dependencyMap, *it, idx, n, oplus, ordersmall))
			) {
				map< pair<signed char, signed char>, set<signed char> > assocCheckMapCopy = assocCheckMap;
				if (testAssocCheckMap(assocCheckMapCopy, *it, idx, n, oplus, perm_left, perm_right, ordersmall)) {
#ifdef DEBUG
					cerr << "  Assoc Check Map: " << endl;
					for (map< pair<signed char, signed char>, set<signed char> >::const_iterator itassoc = assocCheckMapCopy.begin(); itassoc != assocCheckMapCopy.end(); itassoc++) {
						cerr << "    (" << (int)((*itassoc).first).first << ", " << (int)((*itassoc).first).second << "): ";
						for (set<signed char>::const_iterator itset = ((*itassoc).second).begin(); itset != ((*itassoc).second).end(); itset++)
							cerr << (int) *itset << ", ";
						cerr << endl;
	}
#endif

					oplus[it->x][it->y] = idx;
					cur_state[cur_depth] = idx;
					set<todoElem>::iterator itpom2  = it;
					++itpom2;
					gen_oplus_from_set_todo_from_state_to_state_with_depMap_with_assocCheckMap(itpom2, n, set_todo, ordersmall, oplus, perm_left, perm_right, setWPES, setISOS, cur_depth + 1, from_state, from_was_used, to_state, to_was_used, dependencyMap, cur_state, assocCheckMapCopy);
					oplus[it->x][it->y] = -1;
					cur_state[cur_depth] = 0;
				}
			}
		}
	}
}

void gen_oplus_from_set_todo_from_state_to_state_with_depMap(set<todoElem>::iterator it, signed char n, set<todoElem> &set_todo, int ordersmall[MAX][MAX], signed char oplus[MAX][MAX], signed char perm_left[MAX], signed char perm_right[MAX], set<struct wprepea> &setWPES, set<struct wprepea> &setISOS, int cur_depth, int *from_state, bool *from_was_used, int *to_state, bool *to_was_used, const map<todoElem, set<todoElem> > &dependencyMap, int *cur_state) {

	if (it == set_todo.end()) {
#ifdef TESTOUT
	cerr << "\t\tTesting the state: ";
	set<todoElem>::iterator itertmp = set_todo.begin();
	while (itertmp != set_todo.end()) {
		cerr << (int)oplus[itertmp->x][itertmp->y] << ", ";
		++itertmp;
	}
	cerr << endl;
#endif
		if (no_of_iterations == ITERCOUNT) {
			time_t tim = time(NULL);
			cerr << "[" << tim << "]" << endl;
			no_of_iterations = 0;
			cerr << "\tState: ";
			for (set<todoElem>::iterator itp = set_todo.begin(); itp != set_todo.end(); ++itp)
				cerr << setw(2) << (int)oplus[itp->x][itp->y];
			cerr << endl;	
		}
		no_of_iterations++;

		if (is_weak_pre_pea(n, oplus, perm_right, perm_left)) {

			struct wprepea wp(n, oplus, perm_left, perm_right);
			set<struct wprepea>::iterator it1, it2;

			it1 = setWPES.find(wp);

			it2 = setISOS.find(wp);

			if (it1 == setWPES.end() && it2 == setISOS.end()) {

				setWPES.insert(wp);

				print_op(cout, n, oplus);

				cout << "Left orthosupplement:"<<endl;
				for (int i = 0; i < n; i++)
					cout << setw(3) << (int)perm_left[i];
				cout << endl;

				cout << "Right orthosupplement:"<<endl;
				for (int i = 0; i < n; i++)
					cout << setw(3) << (int)perm_right[i];
				cout << endl << endl;
				
#ifdef DEBUG
				cout << "Storing into WPES set." << endl;
#endif

				
				if (condition(n, oplus))
					cout << "Pre pseudo effect algebra." << endl;
				cout << endl;	
				
				update_ISOS(ordersmall, wp, setWPES, setISOS);
			}


		}
	}	
	else {

		signed char atleast = max(it->x, it->y), 
		    atmost = n - 1, start_from, end_to;

		if (!from_was_used[cur_depth]) {
			start_from = from_state[cur_depth];
			from_was_used[cur_depth] = true;
		}
		else {
			start_from = atleast;
		}

		if (!to_was_used[cur_depth]) {

			bool is_to_state_prefix = true;
			for (int i = 0; i < cur_depth; i++)
				if (cur_state[i] != to_state[i]) {
					is_to_state_prefix = false;
					break;

				}

			if (is_to_state_prefix) {
				end_to = to_state[cur_depth];
				to_was_used[cur_depth] = true;
			}
			else 
				end_to = atmost;
		}
		else 
			end_to = atmost;

		#ifdef DEBUG
		//		cout << "Level: " << level << endl << "At least: " << atleast << endl << "At most: " << atmost << endl;
		#endif

		for (int idx = start_from ; idx <= end_to; idx++) {
			if (idx != it->x && idx != it->y 
 				//(atleast == n - 1 || idx == n - 1 || ordersmall[atleast - 1][idx - 1]) 
				&& (idx == n - 1 || ordersmall[it->x - 1][idx - 1]) 
				&& (idx == n - 1 || ordersmall[it->y - 1][idx - 1])
				&& (idx == n - 1 || isFromUpperBound(dependencyMap, *it, idx, n, oplus, ordersmall))
			) {
				oplus[it->x][it->y] = idx;
				cur_state[cur_depth] = idx;
				set<todoElem>::iterator itpom2  = it;
				++itpom2;
				gen_oplus_from_set_todo_from_state_to_state_with_depMap(itpom2, n, set_todo, ordersmall, oplus, perm_left, perm_right, setWPES, setISOS, cur_depth + 1, from_state, from_was_used, to_state, to_was_used, dependencyMap, cur_state);
				oplus[it->x][it->y] = -1;
				cur_state[cur_depth] = 0;
			}
		}
	}
}

void print_set_todo(set<todoElem> &set_todo) {
	cerr << "n: " << set_todo.size() << endl;
	for (set<todoElem>::iterator it = set_todo.begin(); it != set_todo.end(); ++it) 
		cerr << "(" << (int)it->x << "," << (int)it->y << "), ";
	
	cerr << endl;
}


void test_current_permutation_set_from_state_to_state_depMap_AssocCheckMap(int n, signed char perm[MAX], int ordersmall[MAX][MAX], set<struct wprepea> &setWPES, set<struct wprepea> &setISOS, set<setpermElem> &setPERM, int from_state_length, int *from_state, bool *from_was_used, int to_state_length, int *to_state, bool *to_was_used, int *cur_state) {

	signed char right_ortho[MAX], oplus[MAX][MAX];

	compute_inverse_perm(n, perm, right_ortho);

	set<todoElem> test2;


	if (test_order_ortho(n, ordersmall, perm, right_ortho)) {
		if (init_set_todo(n, test2, ordersmall, perm, right_ortho) &&
		    setPERM.find(setpermElem(n,perm)) == setPERM.end()
		) {

			cerr << "Tested permutation: ";
			for (int i = 1; i < n-1;i++)
				cerr << setw(3) << (int)perm[i];
			cerr << endl;

			set_initial_plus(n, oplus);

			init_ones(n, oplus, perm, right_ortho);

			cerr << "Todo elements: ";
			print_set_todo(test2);

			map<todoElem, set<todoElem> > depMap;

			init_dependency_map(test2, ordersmall, depMap);
			cerr << "Dependency map:" << endl;
			display_dependency_map(cerr, depMap);
			cerr << "---" << endl;
			reduce_dependency_map(depMap);
			cerr << "Reduced dependency map:" << endl;
			display_dependency_map(cerr, depMap);
			cerr << "---" << endl;


			signed char perm_iso[MAX];
			for (int i = 0; i < n; i++) perm_iso[i] = i;
			update_set_perm(1, ordersmall, setPERM, n, perm, perm_iso);

#ifndef PLAN

			cerr << "Starting state:      ";
			if (from_state_length > 0) 
				for (int i = 0; i < from_state_length; i++)
					cerr << from_state[i] << ", ";
			else
				cerr << "-";
			cerr << endl;

			cerr << "Ending state prefix: ";
			if (to_state_length > 0) 
				for (int i = 0; i < to_state_length; i++)
					cerr << to_state[i] << ", ";
			else
				cerr << "-";
			cerr << endl;

			time_t tim = time(NULL);
			cerr << "[" << tim << "]" << endl;

			no_of_iterations = 0UL;

			map< pair<signed char, signed char>, set<signed char> > assocCheckMap;
			gen_oplus_from_set_todo_from_state_to_state_with_depMap_with_assocCheckMap(test2.begin(), n, test2, ordersmall, oplus, perm, right_ortho, setWPES, setISOS, 0, from_state, from_was_used, to_state, to_was_used, depMap, cur_state, assocCheckMap);

			tim = time(NULL);
			cerr << "[" << tim << "]" << endl;
			cerr << "\tRemaining count: " << no_of_iterations << endl;
					
#endif
		}
	}
}

void test_current_permutation_set_from_state_to_state_depMap(int n, signed char perm[MAX], int ordersmall[MAX][MAX], set<struct wprepea> &setWPES, set<struct wprepea> &setISOS, set<setpermElem> &setPERM, int from_state_length, int *from_state, bool *from_was_used, int to_state_length, int *to_state, bool *to_was_used, int *cur_state) {

	signed char right_ortho[MAX], oplus[MAX][MAX];

	compute_inverse_perm(n, perm, right_ortho);

	set<todoElem> test2;


	if (test_order_ortho(n, ordersmall, perm, right_ortho)) {
		if (init_set_todo(n, test2, ordersmall, perm, right_ortho) &&
		    setPERM.find(setpermElem(n,perm)) == setPERM.end()
		) {

			cerr << "Tested permutation: ";
			for (int i = 1; i < n-1;i++)
				cerr << setw(3) << (int)perm[i];
			cerr << endl;

			set_initial_plus(n, oplus);

			init_ones(n, oplus, perm, right_ortho);

			cerr << "Todo elements: ";
			print_set_todo(test2);

			map<todoElem, set<todoElem> > depMap;

			init_dependency_map(test2, ordersmall, depMap);
			cerr << "Dependency map:" << endl;
			display_dependency_map(cerr, depMap);
			cerr << "---" << endl;
			reduce_dependency_map(depMap);
			cerr << "Reduced dependency map:" << endl;
			display_dependency_map(cerr, depMap);
			cerr << "---" << endl;


			signed char perm_iso[MAX];
			for (int i = 0; i < n; i++) perm_iso[i] = i;
			update_set_perm(1, ordersmall, setPERM, n, perm, perm_iso);

#ifndef PLAN

			cerr << "Starting state:      ";
			if (from_state_length > 0) 
				for (int i = 0; i < from_state_length; i++)
					cerr << from_state[i] << ", ";
			else
				cerr << "-";
			cerr << endl;

			cerr << "Ending state prefix: ";
			if (to_state_length > 0) 
				for (int i = 0; i < to_state_length; i++)
					cerr << to_state[i] << ", ";
			else
				cerr << "-";
			cerr << endl;

			time_t tim = time(NULL);
			cerr << "[" << tim << "]" << endl;

			no_of_iterations = 0UL;
			gen_oplus_from_set_todo_from_state_to_state_with_depMap(test2.begin(), n, test2, ordersmall, oplus, perm, right_ortho, setWPES, setISOS, 0, from_state, from_was_used, to_state, to_was_used, depMap, cur_state);

			tim = time(NULL);
			cerr << "[" << tim << "]" << endl;
			cerr << "\tRemaining count: " << no_of_iterations << endl;
					
#endif
		}
	}
}

int main(int argc, char* argv[]) {

	int n = N;

	if (argc < 5) {
		cout << "Usage: " << argv[0] << " <ord_num> <permutation> <from_state> <end_state>" <<endl;
		cout << "  <ord_num>: " << VFROM + 1 << "-" << VTO << endl;
		cout << "  <permutation>: comma separated list of length " << N - 2<< endl;
		cout << "  <from_state>: <state_length>:<state>|-" << endl;
		cout << "    - = from beginning" << endl;
		cout << "    <state_length>: the length of state"<< endl;
		cout << "    <state>: comma separated list of values"<< endl;
		cout << "  <end_state>: <state_length>:<state>|-" << endl;
		cout << "    - = from beginning" << endl;
		cout << "    <state_length>: the length of state"<< endl;
		cout << "    <state>: comma separated list of values"<< endl;

		return 1;

	}
	
	int from;

	from = atoi(argv[1]);

#ifdef DEBUG
	cerr << "From: " << from << endl;
#endif

	from--;

	signed char perm[MAX];
	perm[0] = N - 1; perm[N - 1] = 0;

	if (parse_permutation(argv[2], N, perm) && is_permutation(n, perm)) {

		const int maxdepth = 50; // should be enough for now

		bool from_was_used[maxdepth];
		int from_state[maxdepth];
		bool to_was_used[maxdepth];
		int to_state[maxdepth];
		int cur_state[maxdepth];

		for (int i = 0; i < maxdepth; i++) {
			from_was_used[i] = true;
			to_was_used[i] = true;
		}

		int from_state_length;
		if (parse_state(argv[3], maxdepth, N, from_state_length, from_state, from_was_used)) {

#ifdef DEBUG
		cerr << "Read FROM state length: " << from_state_length << endl;
		if (from_state_length > 0) {
			cerr << "Read state: " ;
			for (int i =0; i < from_state_length; i++) {
				cerr << (int) from_state[i] << ", ";
			}
			cerr << endl;
		}
#endif

			int to_state_length;
			if (parse_state(argv[4], maxdepth, N, to_state_length, to_state, to_was_used)) {	

#ifdef DEBUG
				cerr << "Read TO state length: " << to_state_length << endl;
				if (to_state_length > 0) {
					cerr << "Read state: " ;
					for (int i =0; i < to_state_length; i++) {
						cerr << (int) to_state[i] << ", ";
					}
					cerr << endl;
				}
#endif

				int idx = from - order_offset;
				set<struct wprepea> setWPES, setISOS;
				set<setpermElem> setPERM;
				
			//	for (int i = 1; i < n - 1; i++) perm[i] = i;

				cout << "Order n. " << order_offset + idx + 1 << endl;
				cerr << "Order n. " << order_offset + idx + 1 << endl;

			#ifdef PLAN
				for (int a = 0; a < n - 2; a++) {
					for (int b = 0; b < n - 2; b++)
						cerr << setw(2) << ORDERSARRAY[idx][a][b];
					cerr << endl;
				}
			#endif


				test_current_permutation_set_from_state_to_state_depMap_AssocCheckMap(n, perm, ORDERSARRAY[idx], setWPES, setISOS, setPERM, from_state_length, from_state, from_was_used, to_state_length, to_state, to_was_used, cur_state);
			}
		}

	}

	return 0;
}
