#include <time.h>
#include <stdlib.h>
#include <iostream>
#include <iomanip>
#include <set>

#include "consts.h"
#include "wprepea.h"
#include "wprepea_utils.h"
#include "defs.h"


#define ITERCOUNT 100000000UL
#undef DEBUG

using namespace std;

// global
unsigned long int no_of_iterations = 1;

bool exists_left_orthosupplement(signed char n, signed char op_plus[MAX][MAX]);
bool exists_right_orthosupplement(signed char n, signed char op_plus[MAX][MAX]);
bool condition(int n, int op_plus[MAX][MAX]);
void generate_all_orthosupplements(signed char n, signed char op_plus[MAX][MAX], signed char left_ortho[MAX], signed char right_ortho[MAX]);
//bool is_defined(const int op_plus[MAX][MAX], int elem_a, int elem_b);
void dump_state(signed char n, signed char op_plus[MAX][MAX]); 
inline bool is_defined(const signed char op_plus[MAX][MAX], int elem_a, int elem_b);
inline int get_oplus(const signed char op_plus[MAX][MAX], int elem_a, int elem_b);
bool is_weak_pre_pea(signed char n, signed char oplus[MAX][MAX], signed char compl_right[MAX], signed char compl_left[MAX]); 
void print_op(int n, const int op[MAX][MAX]); 




//bool operator<(const struct wprepea &wp1, const struct wprepea &wp2); 

struct wprepeaCompare {
	bool operator() (const struct wprepea &wp1, const struct wprepea &wp2) const;
};

/*bool wprepeaCompare::operator() (const struct wprepea &wp1, const struct wprepea &wp2) const {
	return operator<(wp1, wp2);
}
*/





inline bool is_defined(const signed char op_plus[MAX][MAX], int elem_a, int elem_b) {
	return (get_oplus(op_plus, elem_a, elem_b) != -1);
}

inline int get_oplus(const signed char op_plus[MAX][MAX], int elem_a, int elem_b) {
	return op_plus[elem_a][elem_b];	
}



bool are_left_right_orders_equal(signed char n, signed char op_plus[MAX][MAX], signed char left_ortho[MAX], signed char right_ortho[MAX]) {
	for (int elem_a = 0; elem_a < n; elem_a++)
	for (int elem_b = 0; elem_b < n; elem_b++)
		if ((is_defined(op_plus, elem_a, right_ortho[elem_b]) && !is_defined(op_plus, left_ortho[elem_b], elem_a)) ||
			(is_defined(op_plus, left_ortho[elem_b], elem_a) && !is_defined(op_plus, elem_a, right_ortho[elem_b]))
		)
			return false;

	return true;
}

void dump_state(signed char n, signed char op_plus[MAX][MAX]) {
	cerr << "OP: ";
	for (int i = 1; i < n-1; i++)
		for (int j = 1; j < n-1;j++)
			cerr << (int)op_plus[j][i] << ", " ;
	cerr << endl;
}

bool test_partial_order_right(signed char n, signed char op_plus[MAX][MAX], signed char right_ortho[MAX]) {


	// if a + b~ and b + a~ are defined and a != b
	// antisymmetry
	for (int elem_a = 1; elem_a < n - 1; elem_a++)
	for (int elem_b = 1; elem_b < n - 1; elem_b++)
		if (is_defined(op_plus, elem_a, right_ortho[elem_b]) &&
			is_defined(op_plus, elem_b, right_ortho[elem_a]) &&
			elem_a != elem_b
			)
			return false;

	// transitivity
	// a R b and b R c => a R c
	for (int elem_a = 1; elem_a < n - 1; elem_a++)
	for (int elem_b = 1; elem_b < n - 1; elem_b++)
	for (int elem_c = 1; elem_c < n - 1; elem_c++)
		if ( is_defined(op_plus, elem_a, right_ortho[elem_b]) &&
		     is_defined(op_plus, elem_b, right_ortho[elem_c]) &&
		    !is_defined(op_plus, elem_a, right_ortho[elem_c])
		)
			return false;

	return true;
}

bool test_partial_order_left(signed char n, signed char op_plus[MAX][MAX], signed char left_ortho[MAX]) {

	// antisymmetry
	for (int elem_a = 1; elem_a < n - 1; elem_a++)
	for (int elem_b = 1; elem_b < n - 1; elem_b++)
		if (is_defined(op_plus, left_ortho[elem_b], elem_a) &&
			is_defined(op_plus, left_ortho[elem_a], elem_b) &&
			elem_a != elem_b
			) {
			cout << "Left order Antisymmetry failed for: " << elem_a << ", " << elem_b << "." << endl;
			return false;
		}

	// transitivity
	for (int elem_a = 1; elem_a < n - 1; elem_a++)
	for (int elem_b = 1; elem_b < n - 1; elem_b++)
	for (int elem_c = 1; elem_c < n - 1; elem_c++)
		if ( is_defined(op_plus, left_ortho[elem_b], elem_a) &&
			 is_defined(op_plus, left_ortho[elem_c], elem_b) &&
			!is_defined(op_plus, left_ortho[elem_c], elem_a)
		) {
			cout << "Left order Transitivity failed for: " << elem_a << ", " << elem_b << ", " << elem_c << "." << endl;
			return false;
		}

#ifdef DEBUG
	cout << "Left order Reflexivity: OK." << endl;
	cout << "Left order Antisymmetry: OK." << endl;
	cout << "Left order Transitivity: OK." << endl;
#endif
	return true;
}


bool test_associativity(signed char n, signed char op_plus[MAX][MAX]) {
	for (int a = 1; a < n - 1; a++)
		for (int b = 1; b < n - 1; b++)
			for (int c = 1; c < n - 1; c++) {

				if (is_defined(op_plus, a, b) && is_defined(op_plus, get_oplus(op_plus, a, b), c)) {
					if (!is_defined(op_plus, b, c) || !is_defined(op_plus, a, get_oplus(op_plus, b, c))){
#ifdef DEBUG
						cout << "Associativity failed for:" << a << ", " << b << ", " << c << "." << endl;
#endif
						return false;
					}
					if (get_oplus(op_plus, get_oplus(op_plus, a, b), c) != 
					    get_oplus(op_plus, a, get_oplus(op_plus, b, c))) {

#ifdef DEBUG
						cout << "Associativity failed for:" << a << ", " << b << ", " << c << "." << endl;
#endif
						return false;
					}
				}

				if (is_defined(op_plus, b, c) && is_defined(op_plus, a, get_oplus(op_plus, b, c))) {
					if (!is_defined(op_plus, a, b) || !is_defined(op_plus, get_oplus(op_plus, a, b), c)) {
#ifdef DEBUG
						cout << "Associativity failed for:" << a << ", " << b << ", " << c << "." << endl;
#endif
						return false;
					}
					if (get_oplus(op_plus, get_oplus(op_plus, a, b), c) != 
					    get_oplus(op_plus, a, get_oplus(op_plus, b, c))) {
#ifdef DEBUG

						cout << "Associativity failed for:" << a << ", " << b << ", " << c << "." << endl;
#endif
						return false;
					}
				}
			}
#ifdef DEBUG
	cout << "Associativity: OK." << endl;
#endif
	return true;
}


void print_op(signed char n, const signed char op[MAX][MAX]) {
	for (int i = 0; i < n; i++) {
		for (int j = 0; j < n; j++) {

			if (is_defined(op, i, j))
				cout << setw(3) << (int)get_oplus(op, i, j);
			else
				cout << setw(3) << ".";
		}
		cout << endl;
	}
}

void generate_op_plus_recursive(int level_x, int level_y, signed char n, signed char op_plus[MAX][MAX]) {

	if (level_y == n - 1) {
		if (no_of_iterations == ITERCOUNT) {
			no_of_iterations = 0;
			dump_state(n, op_plus);
		}
		no_of_iterations++;

		// Test properties
		if (	
			test_associativity(n, op_plus) &&
			exists_left_orthosupplement(n, op_plus)  &&
			exists_right_orthosupplement(n, op_plus) /*&&
			condition(n, op_plus) */
		) {
			//print_op(n, op_plus);

			signed char left_ortho[MAX], right_ortho[MAX];

			generate_all_orthosupplements(n, op_plus, left_ortho, right_ortho);
		}
	}
	else {
		if (level_x == 1 && level_y == 1) {
			for (int elem_a = -1; elem_a <= n / 2; elem_a++) {
				if (elem_a != 0 && elem_a != level_x && elem_a != level_y) {

					set_oplus(op_plus, level_x, level_y, elem_a);

					if (level_x < n - 2)
						generate_op_plus_recursive(level_x + 1, level_y, n, op_plus);
					else
						generate_op_plus_recursive(1, level_y + 1, n, op_plus);

					set_oplus(op_plus, level_x, level_y, -1);
				}
			}
		}
		else {
			for (int elem_a = -1; elem_a < n; elem_a++) {
				if (elem_a != 0 && elem_a != level_x && elem_a != level_y) {

					set_oplus(op_plus, level_x, level_y,  elem_a);

					if (level_x < n - 2)
						generate_op_plus_recursive(level_x + 1, level_y, n, op_plus);
					else
						generate_op_plus_recursive(1, level_y + 1, n, op_plus);

					set_oplus(op_plus, level_x, level_y, -1);
				}
			}
		}
	}
}

bool exists_left_orthosupplement(signed char n, signed char op_plus[MAX][MAX]) {
	for (int i = 0; i < n; i++) {
		bool found = false;

		for (int j = 0; j < n; j++)
			if (get_oplus(op_plus, j, i) == n - 1) {
				found = true;
				break;
			}
		if (!found)
			return false;
	}
	return true;
}

bool exists_right_orthosupplement(signed char n, signed char op_plus[MAX][MAX]) {
	for (int i = 0; i < n; i++) {
		bool found = false;

		for (int j = 0; j < n; j++)
			if (get_oplus(op_plus, i, j) == n - 1) {
				found = true;
				break;
			}
		if (!found)
			return false;
	}
	return true;
}

bool condition(signed char n, signed char op_plus[MAX][MAX]) {
	for (int i = 0; i < n; i++)
	for (int j = 0; j < n; j++)
		if (is_defined(op_plus, i, j)) {
			bool found_c = false;
			int result = get_oplus(op_plus, i, j);

			for (int k = 0; k < n; k++)
				if (get_oplus(op_plus, k, i) == result) {
					found_c = true;
					break;
				}
			if (!found_c)
				return false;

			bool found_d = false;


			for (int k = 0; k < n; k++)
				if (get_oplus(op_plus, j, k) == result) {
					found_d = true;
					break;
				}
			if (!found_d)
				return false;
		}
	return true;
}

void generate_all_orthosupplements_left_recursive(int level, signed char n, signed char op_plus[MAX][MAX], signed char left_ortho[MAX], signed char right_ortho[MAX]) {

	if (level == n - 1) {



		bool left_order, right_order;

		// Test partial order left
		left_order = test_partial_order_left(n, op_plus, left_ortho);

		// Test partial order right
		right_order = test_partial_order_right(n, op_plus, right_ortho);

		//print_op(n, op_plus);

		bool equal_left_right_order = left_order && right_order && are_left_right_orders_equal(n, op_plus, left_ortho, right_ortho);

		if (equal_left_right_order) {

			print_op(n, op_plus);

			cout << "Left orthosupplement:"<<endl;
			for (int i = 0; i < n; i++)
				cout << setw(3) << (int)left_ortho[i];
			cout << endl;

			cout << "Right orthosupplement:"<<endl;
			for (int i = 0; i < n; i++)
				cout << setw(3) << (int)right_ortho[i];
			cout << endl;



		/*	cout << "Flags: ";
			if (left_order)
				cout << "L";
			else
				cout << ".";

			if (right_order)
				cout << "R";
			else
				cout << ".";

			if (left_order && right_order && are_left_right_orders_equal(n, op_plus, left_ortho, right_ortho))
				cout << "=";
			else
				cout << ".";
		*/

			cout << endl;
		}

	}
	else {
		for (int k = 0; k < n; k++)
			if (op_plus[k][level] == n - 1) {
				left_ortho[level] = k;
				generate_all_orthosupplements_left_recursive(level + 1, n, op_plus, left_ortho, right_ortho);
			}
	}


}
void generate_all_orthosupplements_right_recursive(int level, signed char n, signed char op_plus[MAX][MAX], signed char left_ortho[MAX], signed char right_ortho[MAX]) {

	if (level == n - 1) {

		//generate_all_orthosupplements_left_recursive(1, n, op_plus, left_ortho, right_ortho);

		// compute left orthosupplement by right orthosupplement
		// i.e. compute inverse permutation the the right orthosupplement
		for (int k = 1; k < n-1; k++) {
			left_ortho[right_ortho[k]] = k;
		}

		bool left_order, right_order;

		// Test partial order left
		left_order = test_partial_order_left(n, op_plus, left_ortho);

		// Test partial order right
		right_order = test_partial_order_right(n, op_plus, right_ortho);

		//print_op(n, op_plus);

		bool equal_left_right_order = left_order && right_order && are_left_right_orders_equal(n, op_plus, left_ortho, right_ortho);

		if (equal_left_right_order) {

			print_op(n, op_plus);

			cout << "Left orthosupplement:"<<endl;
			for (int i = 0; i < n; i++)
				cout << setw(3) << (int)left_ortho[i];
			cout << endl;

			cout << "Right orthosupplement:"<<endl;
			for (int i = 0; i < n; i++)
				cout << setw(3) << (int)right_ortho[i];
			cout << endl;



		/*	cout << "Flags: ";
			if (left_order)
				cout << "L";
			else
				cout << ".";

			if (right_order)
				cout << "R";
			else
				cout << ".";

			if (left_order && right_order && are_left_right_orders_equal(n, op_plus, left_ortho, right_ortho))
				cout << "=";
			else
				cout << ".";
		*/

			cout << endl;
		}


	}
	else {
		for (int k = 1; k < n - 1; k++)
			if (get_oplus(op_plus, level, k) == n - 1) {
				bool found = false;
				
				for (int i = 1; i < level; i++)
					if (right_ortho[i] == k) {
						found = true;
						break;
					}
						
				if (!found) {	
					right_ortho[level] = k;
					generate_all_orthosupplements_right_recursive(level + 1, n, op_plus, left_ortho, right_ortho);
				}
			}
	}

}


void generate_all_orthosupplements(signed char n, signed char op_plus[MAX][MAX], signed char left_ortho[MAX], signed char right_ortho[MAX]) {
	left_ortho[0] = n - 1;
	left_ortho[n - 1] = 0;
	right_ortho[0] = n - 1;
	right_ortho[n - 1] = 0;

	generate_all_orthosupplements_right_recursive(1, n, op_plus, left_ortho, right_ortho);
}



void generate_op_plus(signed char n, signed char op_plus[MAX][MAX]) {
	set_initial_plus(n, op_plus);

	generate_op_plus_recursive(1, 1, n, op_plus);
}

struct todo {
	int n;
	int pairs[MAX*MAX][2];
} TODO;

struct todo test1 = {7, {{1, 1}, {1, 2}, {1, 5}, {2, 1},  {2, 2}, {4,2}, {5, 2}}};
int perm_left1[MAX] = {6,5,3,1,2,4,0};
int perm_right1[MAX]= {6,3,4,2,5,1,0};

//int ordersmall[MAX][MAX] = {
//	{1,0,1,1,0},{0,1,1,1,1},{0,0,1,0,0},{0,0,0,1,0},{0,0,0,0,1}	
//	{1,0,0,0,0},{0,1,0,0,0},{0,0,1,0,0},{0,0,0,1,0},{0,0,0,0,1}	
//};



void init_ones(signed char n, signed char oplus[MAX][MAX], signed char perm_left[MAX], signed char perm_right[MAX]) {

	cerr << "Fixed: ";
	for (int idx = 1; idx < n - 1; idx++) {
		oplus[idx][perm_right[idx]] = n - 1;
		cerr << "(" << idx << "," << (int)perm_right[idx] << "), ";
	}
	cerr << endl;
}





/*
bool operator<(const struct todoElem &te1, const struct todoElem &te2) {
	if (te1.x < te2.x || (te1.x == te2.x && te1.y < te2.y))
		return true;

	return false;
}
*/


void gen_oplus_from_todo(int level, signed char n, struct todo todo_list, int ordersmall[MAX][MAX], signed char oplus[MAX][MAX], signed char perm_left[MAX], signed char perm_right[MAX], set<struct wprepea> &setWPES, set<struct wprepea> &setISOS) {

	if (level == todo_list.n) {
		if (is_weak_pre_pea(n, oplus, perm_right, perm_left)) {

			struct wprepea wp(n, oplus, perm_left, perm_right);
			set<struct wprepea>::iterator it1, it2;

			it1 = setWPES.find(wp);

			it2 = setISOS.find(wp);

			if (it1 == setWPES.end() && it2 == setISOS.end()) {

				setWPES.insert(wp);

				print_op(n, oplus);

				cout << "Left orthosupplement:"<<endl;
				for (int i = 0; i < n; i++)
					cout << setw(3) << (int)perm_left[i];
				cout << endl;

				cout << "Right orthosupplement:"<<endl;
				for (int i = 0; i < n; i++)
					cout << setw(3) << (int)perm_right[i];
				cout << endl << endl;
				
#ifdef DEBUG
				cout << "Storing into WPES set." << endl;
#endif

				if (condition(n, oplus))
					cout << "Pre pseudo effect algebra." << endl;
				
/*
				int isoperm[MAX];
				for (int i = 0; i < n ; i++)
					isoperm[i] = i;

				update_ISOS(1, ordersmall, wp, isoperm, setWPES, setISOS);
*/
				update_ISOS(ordersmall, wp, setWPES, setISOS);
			}


		}
	}	
	else {

		signed char atleast = max(todo_list.pairs[level][0], todo_list.pairs[level][1]), 
		    atmost = n - 1;
		for (int idx = 0; idx < level; idx++) {
			if (ordersmall[todo_list.pairs[idx][0] - 1]
				      [todo_list.pairs[level][0] - 1] 
				&& 
			    ordersmall[todo_list.pairs[idx][1] -1]
				      [todo_list.pairs[level][1] - 1]
			   ) {
				atleast = max(atleast, oplus[todo_list.pairs[idx][0]][todo_list.pairs[idx][1]]);

			}

			if (ordersmall[todo_list.pairs[level][0] - 1]
				      [todo_list.pairs[idx][0] - 1] 
				&& 
			    ordersmall[todo_list.pairs[level][1] -1]
				      [todo_list.pairs[idx][1] - 1]
			   ) {
				atmost = min(atmost, oplus[todo_list.pairs[idx][0]][todo_list.pairs[idx][1]]);

			}


		}

#ifdef DEBUG
		cout << "Level: " << level << endl << "At least: " << atleast << endl << "At most: " << atmost << endl;
#endif

		for (int idx = atleast ; idx <= atmost; idx++) {
			if (/*idx != 0 && */idx != todo_list.pairs[level][0] && idx != todo_list.pairs[level][1] &&
				atleast <= atmost && 
				(atleast == n - 1 || idx == n - 1 || ordersmall[atleast - 1][idx - 1]) &&
			        (idx == n - 1 || atmost == n - 1 || ordersmall[idx - 1][atmost - 1])
			) {
				oplus[todo_list.pairs[level][0]][todo_list.pairs[level][1]] = idx;
				gen_oplus_from_todo(level + 1, n, todo_list, ordersmall, oplus, perm_left, perm_right, setWPES, setISOS);
				oplus[todo_list.pairs[level][0]][todo_list.pairs[level][1]] = -1;
			}
		}
	}
}

void gen_oplus_from_set_todo(set<todoElem>::iterator it, signed char n, set<todoElem> &set_todo, int ordersmall[MAX][MAX], signed char oplus[MAX][MAX], signed char perm_left[MAX], signed char perm_right[MAX], set<struct wprepea> &setWPES, set<struct wprepea> &setISOS) {

	if (it == set_todo.end()) {
		if (no_of_iterations == ITERCOUNT) {
			time_t tim = time(NULL);
			cerr << "[" << tim << "]" << endl;
			no_of_iterations = 0;
			cerr << "\tState: ";
			for (set<todoElem>::iterator itp = set_todo.begin(); itp != set_todo.end();itp++)
				cerr << setw(2) << (int)oplus[itp->x][itp->y];
			cerr << endl;	
		}
		no_of_iterations++;

		if (is_weak_pre_pea(n, oplus, perm_right, perm_left)) {

			struct wprepea wp(n, oplus, perm_left, perm_right);
			set<struct wprepea>::iterator it1, it2;

			it1 = setWPES.find(wp);

			it2 = setISOS.find(wp);

			if (it1 == setWPES.end() && it2 == setISOS.end()) {

				setWPES.insert(wp);

				print_op(n, oplus);

				cout << "Left orthosupplement:"<<endl;
				for (int i = 0; i < n; i++)
					cout << setw(3) << (int)perm_left[i];
				cout << endl;

				cout << "Right orthosupplement:"<<endl;
				for (int i = 0; i < n; i++)
					cout << setw(3) << (int)perm_right[i];
				cout << endl;
				
#ifdef DEBUG
				cout << "Storing into WPES set." << endl;
#endif

				if (condition(n, oplus))
					cout << "Pre pseudo effect algebra." << endl;
				cout << endl;	

/*
				int isoperm[MAX];
				for (int i = 0; i < n ; i++)
					isoperm[i] = i;

				update_ISOS(1, ordersmall, wp, isoperm, setWPES, setISOS);
*/
				update_ISOS(ordersmall, wp, setWPES, setISOS);
			}


		}
	}	
	else {

		signed char atleast = max(it->x, it->y), 
		    atmost = n - 1;
		for (set<todoElem>::iterator itpom = set_todo.begin(); itpom != it; itpom++) {
			if (ordersmall[itpom->x - 1]
				      [it->x - 1] 
				&& 
			    ordersmall[itpom->y -1]
				      [it->y - 1]
			   ) {
				atleast = max(atleast, oplus[itpom->x][itpom->y]);

			}

#if 0
			if (ordersmall[it->x - 1]
				      [itpom->x - 1] 
				&& 
			    ordersmall[it->y -1]
				      [itpom->y - 1]
			   ) {
				atmost = min(atmost, oplus[itpom->x][itpom->y]);

			}
#endif


		}

#ifdef DEBUG
//		cout << "Level: " << level << endl << "At least: " << atleast << endl << "At most: " << atmost << endl;
#endif

		for (int idx = atleast ; idx <= atmost; idx++) {
			if (/*idx != 0 && */idx != it->x && idx != it->y &&
				/*atleast <= atmost && */
 				(atleast == n - 1 || idx == n - 1 || ordersmall[atleast - 1][idx - 1]) /*&&
			        (idx == n - 1 || atmost == n - 1 || ordersmall[idx - 1][atmost - 1])*/
		&& (idx == n - 1 || ordersmall[it->x - 1][idx - 1]) 
		&& (idx == n - 1 || ordersmall[it->y - 1][idx - 1])
			) {
				oplus[it->x][it->y] = idx;
				set<todoElem>::iterator itpom2  = it;
				itpom2++;
				gen_oplus_from_set_todo(itpom2, n, set_todo, ordersmall, oplus, perm_left, perm_right, setWPES, setISOS);
				oplus[it->x][it->y] = -1;
			}
		}
	}
}

bool is_weak_pre_pea(signed char n, signed char oplus[MAX][MAX], signed char compl_right[MAX], signed char compl_left[MAX]) {
	
	if (	test_associativity(n, oplus) /*&& 
		test_partial_order_right(n, oplus, compl_right)*/ /*&& 
		test_partial_order_left(n, oplus, compl_left) &&
		are_left_right_orders_equal(n, oplus, compl_left, compl_right)*/
	)
		return true;

	return false;
}

bool search_in_todo(struct todo todo_list, int pair_x, int pair_y) {
	for (int i = 0; i < todo_list.n;i++)
		if (todo_list.pairs[i][0]==pair_x && todo_list.pairs[i][1]==pair_y)
			return true;
	return false;
}

void insert_into_todo(struct todo &todo_list, int pair_x, int pair_y) {
	
	if (!search_in_todo(todo_list, pair_x, pair_y)) {
		int idx = todo_list.n;
		todo_list.pairs[idx][0] = pair_x;
		todo_list.pairs[idx][1] = pair_y;
		todo_list.n++;
	}
}

void print_todo(struct todo todo_list) {
	cout << "n: " << todo_list.n << endl;
	for (int i = 0; i < todo_list.n; i++)
		cout << "(" << todo_list.pairs[i][0] << "," << todo_list.pairs[i][1] << "), ";
	cout << endl;
}

void print_set_todo(set<todoElem> &set_todo) {
	cerr << "n: " << set_todo.size() << endl;
	for (set<todoElem>::iterator it = set_todo.begin(); it != set_todo.end(); it++) 
		cerr << "(" << (int)it->x << "," << (int)it->y << "), ";
	
	cerr << endl;
}

bool search_todo(const struct todo &todo_list, int a, int b) {
	for (int i = 0; i < todo_list.n; i++)
		if (todo_list.pairs[i][0]==a && todo_list.pairs[i][1]==b)
			return true;
	
	return false;
}


bool init_todo(signed char n, struct todo &todo_list, int ordersmall[MAX][MAX], signed char left_ortho[MAX], signed char right_ortho[MAX]) {
	todo_list.n = 0;

	for (int i = 0; i < n - 2; i++)
	for (int j = i+1; j < n - 2; j++) {
		if (ordersmall[i][j]) {
			//insert (i, rightortho[j]), (leftortho[j], i) into todo_list

			insert_into_todo(todo_list, i + 1, right_ortho[j + 1]);
			insert_into_todo(todo_list, left_ortho[j + 1], i + 1);
		}
	}

	for (int i = 0; i < n - 2; i++)
	for (int j = 0; j < n - 2; j++) {
		if (i != j && !ordersmall[i][j]) {

			if (search_todo(todo_list, i + 1, right_ortho[j + 1]))
				return false;

			if (search_todo(todo_list, left_ortho[j + 1], i + 1))
				return false;

			if (search_todo(todo_list, left_ortho[left_ortho[i + 1]], left_ortho[j + 1]))
				return false;

			if (search_todo(todo_list, right_ortho[j + 1], right_ortho[right_ortho[i + 1]]))
				return false;
			
		}
	}


	return true;
}


bool test_order_ortho(signed char n, int ordersmall[MAX][MAX], signed char left_ortho[MAX], signed char right_ortho[MAX]) {
	for (int i = 1; i < n - 2; i++)
		for (int j = i + 1; j < n - 1; j++) 
			if (ordersmall[i - 1][j - 1] && (!ordersmall[left_ortho[j] - 1][left_ortho[i] - 1] || !ordersmall[right_ortho[j] - 1][right_ortho[i] - 1]))
				return false;
			
		
	return true;
}


void gen_all_permutations(int level, signed char n, signed char perm[MAX], int ordersmall[MAX][MAX], set<struct wprepea> &setWPES, set<struct wprepea> &setISOS) {
	int tmp;

	if (level == n-1) {
		signed char right_ortho[MAX], oplus[MAX][MAX];

		compute_inverse_perm(n, perm, right_ortho);

		struct todo test2;

#ifdef DEBUG
		cout << "Current permutation: ";
		for (int i = 1; i < n-1;i++)
			cout << setw(3) << perm[i];
		cout << endl;
#endif
/*
		cout << "-1" << endl;
		for (int i = 1; i < n-1;i++)
			cout << setw(3) << right_ortho[i];
		cout << endl;
*/

//		if (
		
//		for (int idx = 0; idx < order6count; idx++) {

			if (test_order_ortho(n, ordersmall, perm, right_ortho)) {

				set_initial_plus(n, oplus);
	#ifdef DEBUG
				cout << "set_initial_plus" << endl;
				print_op(n, oplus);
	#endif

				init_ones(n, oplus, perm, right_ortho);

	#ifdef DEBUG
				cout << "init_ones" << endl;
				print_op(n, oplus);
	#endif

				if (
				 init_todo(n, test2, ordersmall, perm, right_ortho)
				) {
	#ifdef DEBUG
					print_todo(test2);
	#endif

					gen_oplus_from_todo(0, n, test2, ordersmall, oplus, perm, right_ortho, setWPES, setISOS);
				}
			}
//		}


	}
	else {
		for (int idx=level; idx < n-1; idx++) {
			if (idx != level) {
				tmp = perm[idx];
				perm[idx] = perm[level];
				perm[level] = tmp;
			}
			gen_all_permutations(level+1, n, perm, ordersmall, setWPES, setISOS);

			if (idx != level) {
				tmp = perm[idx];
				perm[idx] = perm[level];
				perm[level] = tmp;
			}
		}
	}	


}

struct setpermElem {
	signed char n;
	signed char perm[MAX];

	setpermElem(signed char _n, signed char _perm[MAX]) {
		n = _n;	
		for (int i = 0; i < n; i++)
			perm[i] = _perm[i];
	}

	friend bool operator<(const struct setpermElem &spe1, const struct setpermElem &spe2);
};

bool operator<(const struct setpermElem &spe1, const struct setpermElem &spe2) {
	if (spe1.n < spe2.n)
		return true;
	if (spe1.n > spe2.n)
		return false;

	for (int i = 0; i < spe1.n ; i++)
		if (spe1.perm[i] != spe2.perm[i]) {
			if (spe1.perm[i] < spe2.perm[i]) 
				return true;
			else 
				return false;
		}

	return false;
}



bool preserve_order(int n, int ordersmall[MAX][MAX], signed char perm[MAX]) {
	for (int ii=1; ii < n - 1; ii++)
	for (int jj=1; jj < n  -1; jj++)
		if (ii != jj) {
			if (
				(ordersmall[ii - 1][jj - 1] && !ordersmall[perm[ii] - 1][perm[jj] -1]) ||
				(!ordersmall[ii - 1][jj - 1] && ordersmall[perm[ii] - 1][perm[jj] - 1])
			) {
				return false;
			}
		}
	return true;
}


void update_set_perm(int level, int ordersmall[MAX][MAX], set<setpermElem> &setPERM, int n, signed char perm[MAX], signed char perm_iso[MAX]) {

	if (level== n - 1) {
		
		signed char applied_perm[MAX];

		applied_perm[0] = n - 1;	
		applied_perm[n - 1] = 0;	

		for (int i = 1; i < n - 1; i++) 
			applied_perm[perm_iso[i]]  = perm_iso[perm[i]];	


		if (is_admissible(n, ordersmall, applied_perm)) {

			if (preserve_order(n, ordersmall, perm_iso)) {

#ifdef DEBUG
		cerr << "Perm ISO: ";
		for (int i = 1; i < n - 1; i++)
			cerr << setw(2) << (int) perm[i] ;
		cerr <<endl;
		cerr << "Applied Perm: ";
		for (int i = 1; i < n - 1; i++)
			cerr << setw(2) << (int) applied_perm[i] ;
		cerr <<endl;
#endif
#if 0
				cerr << "Is in set?: " ;
				for (int i = 1; i < n - 1; i++) {
					cerr << setw(3) << (int) applied_perm[i];
					
				}
				cerr << endl;
#endif

#if 0
		for (int i=1; i < n - 1; i++)
		for (int j=1; j < n  -1; j++)
			if (i != j) {
				if (
					(ordersmall[i - 1][j - 1] && !ordersmall[perm[i] - 1][perm[j] -1]) ||
					(!ordersmall[i - 1][j - 1] && ordersmall[perm[i] - 1][perm[j] - 1])
				)
					return false;
		}
		return true; 
	
#endif 
		


				if (setPERM.find(setpermElem(n, applied_perm)) == setPERM.end()) {

	#ifdef PLAN
					cerr << "\t Isomorphic perm: " ;
					for (int i = 1; i < n - 1; i++) {
						cerr << setw(3) << (int) applied_perm[i];
						
					}
					cerr << " ("; 
					for (int i = 1; i < n - 1; i++) {
						cerr << setw(2) << (int) perm_iso[i];
						
					}
					cerr << ")";

					cerr << endl;
	#endif

					setPERM.insert(setpermElem(n, applied_perm));
				}
	//			cerr << "setPERM size: " << setPERM.size() << endl;
			}
		}

	}
	else {
		for (int i = level; i < n - 1; i++) {
			if (i != level) {
				signed char tmp = perm_iso[i];
				perm_iso[i] = perm_iso[level];
				perm_iso[level] = tmp;
			}
			update_set_perm(level + 1, ordersmall, setPERM, n, perm, perm_iso);
			if (i != level) {
				signed char tmp = perm_iso[i];
				perm_iso[i] = perm_iso[level];
				perm_iso[level] = tmp;
			}
		}
	}
}



void gen_all_permutations_set(int level, signed char n, signed char perm[MAX], int ordersmall[MAX][MAX], set<struct wprepea> &setWPES, set<struct wprepea> &setISOS, set<setpermElem> &setPERM) {
	int tmp;

	if (level == n-1) {
		signed char right_ortho[MAX], oplus[MAX][MAX];

		compute_inverse_perm(n, perm, right_ortho);

		set<todoElem> test2;

//#ifdef DEBUG
#if 0
		cerr << "Current permutation: ";
		for (int i = 1; i < n-1;i++)
			cerr << setw(3) << (int)perm[i];
		cerr << endl;
#endif
/*
		cout << "-1" << endl;
		for (int i = 1; i < n-1;i++)
			cout << setw(3) << right_ortho[i];
		cout << endl;
*/

//		if (
		
//		for (int idx = 0; idx < order6count; idx++) {

			if (test_order_ortho(n, ordersmall, perm, right_ortho)) {


				if (
				 init_set_todo(n, test2, ordersmall, perm, right_ortho) &&
					setPERM.find(setpermElem(n,perm)) == setPERM.end()
				) {

					cerr << "Tested permutation: ";
					for (int i = 1; i < n-1;i++)
						cerr << setw(3) << (int)perm[i];
					cerr << endl;

						set_initial_plus(n, oplus);
			#ifdef DEBUG
						cout << "set_initial_plus" << endl;
						print_op(n, oplus);
			#endif

						init_ones(n, oplus, perm, right_ortho);

			#ifdef DEBUG
						cout << "init_ones" << endl;
						print_op(n, oplus);
			#endif

					//init_set_todo(n, test2, ordersmall, perm, right_ortho);
	#if 1
					cerr << "Todo elements: ";
					print_set_todo(test2);
	#endif
					signed char perm_iso[MAX];
					for (int i = 0; i < n; i++) perm_iso[i] = i;
					update_set_perm(1, ordersmall, setPERM, n, perm, perm_iso);

#ifndef PLAN
					time_t tim = time(NULL);
					cerr << "[" << tim << "]" << endl;
					no_of_iterations = 1UL;
					gen_oplus_from_set_todo(test2.begin(), n, test2, ordersmall, oplus, perm, right_ortho, setWPES, setISOS);
					tim = time(NULL);
					cerr << "[" << tim << "]" << endl;
					cerr << "\tRemaining count: " << no_of_iterations << endl;
					
#endif
				}
			}
//		}


	}
	else {
		for (int idx=level; idx < n-1; idx++) {
			if (idx != level) {
				tmp = perm[idx];
				perm[idx] = perm[level];
				perm[level] = tmp;

			}
			gen_all_permutations_set(level+1, n, perm, ordersmall, setWPES, setISOS, setPERM);

			if (idx != level) {
				tmp = perm[idx];
				perm[idx] = perm[level];
				perm[level] = tmp;
			}
		}
	}	


}


bool parse_order_specification(char *spec, int &from, int &to) {
	char *pom = spec;

	if (*pom == '-') {
		// no starting range
		from = VFROM + 1;
		pom++;
		if (!*pom) {
			// no end range
			to = VTO;
		} else {
			if (*pom >= '0' && *pom <= '9') {
				// read number to
				int num = 0;
				while (*pom >= '0' && *pom <= '9') {
					num = 10 * num + (*pom - '0');
					pom++;
				}
				to = num;
			}
			else {
				cerr << "Unknown character '" << *pom << "'." << endl;
				return false;
			}
		}
	
	} else {
		if (*pom >= '0' && *pom <= '9') {
			// read number from
			int num = 0;
			while (*pom >= '0' && *pom <= '9') {
				num = 10 * num + (*pom - '0');
				pom++;
			}
			from = num;
			
		}
		else {
			cerr << "Unknown character '" << *pom << "'." << endl;
			return false;
		}
		if (!*pom) {
			//single number
			to = from;
		} else {
			if (*pom == '-') {
				pom++;
				if (!*pom) {
					//range from-DTVO
					to = VTO;
				} else {
					if (*pom >='0' && *pom <= '9') {
						// read number to
						int num = 0;
						while (*pom >= '0' && *pom <= '9') {
							num = 10 * num + (*pom - '0');
							pom++;
						}
						to = num;
					}
					else {
						cerr << "Unknown character '" << *pom << "'." << endl;
						return false;
					}
				}
			}
			else {
				cerr << "Unknown character '" << *pom << "'." << endl;
				return false;
			}

		}
	}


	return true;
}

int main(int argc, char* argv[]) {

	int n = N;

	signed char perm[MAX];
	perm[0] = n - 1; perm[n - 1] = 0;

	if (argc < 2) {
		cout << "Usage: " << argv[0] << " <order_specification>" << endl;
		cout << "\t<order_specification>:"<< endl;
		cout << "\t\t- = test orders from " << VFROM + 1 << " to " << VTO << endl;
		cout << "\t\tord_num- = test orders from ord_num to " << VTO << endl;
		cout << "\t\t-ord_num = test orders from " << VFROM + 1 << "to ord_num" << endl;
		cout << "\t\tord_num = test order ord_num" << endl;
		cout << "\t\tord_num1-ord_num2 = test orders from ord_num1 to ord_num2" << endl;
		return 1;
	}

	int from, to;

	if (parse_order_specification(argv[1], from, to)) {

#ifdef DEBUG
		cout << "from: " << from << ", to: " << to << endl;
#endif

		from--;
		
		if (from < VFROM || to > VTO || from > to) {
			cout << "Allowed range is: " << VFROM + 1 << "-" << VTO << "." << endl;
			return 1;
		}

		for (int idx = (from-order_offset); idx < (to-order_offset); idx++) {
			set<struct wprepea> setWPES, setISOS;
			set<setpermElem> setPERM;
			
			for (int i = 1; i < n - 1; i++) perm[i] = i;

			cout << "Order n. " << order_offset + idx + 1 << endl;
			cerr << "Order n. " << order_offset + idx + 1 << endl;

	#ifdef PLAN
			for (int a = 0; a < n - 2; a++) {
				for (int b = 0; b < n - 2; b++)
					cerr << setw(2) << ORDERSARRAY[idx][a][b];
				cerr << endl;
			}
	#endif


			gen_all_permutations_set(1, n, perm, ORDERSARRAY[idx], setWPES, setISOS, setPERM);
		}
	}

	return 0;
}
