#include <time.h>
#include <stdlib.h>
#include <iostream>
#include <iomanip>
#include <set>

#include "consts.h"
#include "wprepea.h"
#include "wprepea_utils.h"
#include "wprepea_parseutils.h"
#include "defs.h"


#define ITERCOUNT 100000000UL
#undef DEBUG

using namespace std;

// global
unsigned long int no_of_iterations = 0;

void gen_oplus_from_set_todo_with_depMap_with_assocCheckMap(set<todoElem>::iterator it, signed char n, set<todoElem> &set_todo, int ordersmall[MAX][MAX], signed char oplus[MAX][MAX], signed char perm_left[MAX], signed char perm_right[MAX], set<struct wprepea> &setWPES, set<struct wprepea> &setISOS, const map<todoElem, set<todoElem> > &dependencyMap, map< pair<signed char, signed char>, set<signed char> > &assocCheckMap) {

	if (it == set_todo.end()) {
		if (no_of_iterations == ITERCOUNT) {
			time_t tim = time(NULL);
			cerr << "[" << tim << "]" << endl;
			no_of_iterations = 0;
			cerr << "\tState: ";
			for (set<todoElem>::iterator itp = set_todo.begin(); itp != set_todo.end(); ++itp)
				cerr << setw(2) << (int)oplus[itp->x][itp->y];
			cerr << endl;	
		}
		no_of_iterations++;

		if (is_weak_pre_pea(n, oplus, perm_right, perm_left)) {

			struct wprepea wp(n, oplus, perm_left, perm_right);
			set<struct wprepea>::iterator it1, it2;

			it1 = setWPES.find(wp);

			it2 = setISOS.find(wp);

			if (it1 == setWPES.end() && it2 == setISOS.end()) {

				setWPES.insert(wp);

				print_op(cout, n, oplus);

				cout << "Left orthosupplement:"<<endl;
				for (int i = 0; i < n; i++)
					cout << setw(3) << (int)perm_left[i];
				cout << endl;

				cout << "Right orthosupplement:"<<endl;
				for (int i = 0; i < n; i++)
					cout << setw(3) << (int)perm_right[i];
				cout << endl;
				
#ifdef DEBUG
				cout << "Storing into WPES set." << endl;
#endif

				if (condition(n, oplus))
					cout << "Pre pseudo effect algebra." << endl;
				cout << endl;	

				update_ISOS(ordersmall, wp, setWPES, setISOS);
			}


		}
	}	
	else {

		signed char atleast = max(it->x, it->y), 
		    atmost = n - 1;
/*
		for (set<todoElem>::iterator itpom = set_todo.begin(); itpom != it; itpom++) {
			if (ordersmall[itpom->x - 1]
				      [it->x - 1] 
				&& 
			    ordersmall[itpom->y -1]
				      [it->y - 1]
			   ) {
				atleast = max(atleast, oplus[itpom->x][itpom->y]);

			}
*/

#if 0
			if (ordersmall[it->x - 1]
				      [itpom->x - 1] 
				&& 
			    ordersmall[it->y -1]
				      [itpom->y - 1]
			   ) {
				atmost = min(atmost, oplus[itpom->x][itpom->y]);

			}
#endif


	//	}

#ifdef DEBUG
//		cout << "Level: " << level << endl << "At least: " << atleast << endl << "At most: " << atmost << endl;
		cerr << "(" << (int)it->x<< ", " << (int)it->y << "): At least: " << (int) atleast << ", At most: " << (int) atmost << endl;
#endif

		for (int idx = atleast ; idx <= atmost; idx++) {
			if (	idx != it->x && idx != it->y &&
 				(atleast == n - 1 || idx == n - 1 || ordersmall[atleast - 1][idx - 1]) 
				&& (idx == n - 1 || ordersmall[it->x - 1][idx - 1]) 
				&& (idx == n - 1 || ordersmall[it->y - 1][idx - 1])
				&& (idx == n - 1 || isFromUpperBound(dependencyMap, *it, idx, n, oplus, ordersmall))
			) {
				map< pair<signed char, signed char>, set<signed char> > assocCheckMapCopy = assocCheckMap;
				if (testAssocCheckMap(assocCheckMapCopy, *it, idx, n, oplus, perm_left, perm_right, ordersmall)) {
#ifdef DEBUG
					cerr << "  Assoc Check Map: " << endl;
					for (map< pair<signed char, signed char>, set<signed char> >::const_iterator itassoc = assocCheckMapCopy.begin(); itassoc != assocCheckMapCopy.end(); itassoc++) {
						cerr << "    (" << (int)((*itassoc).first).first << ", " << (int)((*itassoc).first).second << "): ";
						for (set<signed char>::const_iterator itset = ((*itassoc).second).begin(); itset != ((*itassoc).second).end(); itset++)
							cerr << (int) *itset << ", ";
						cerr << endl;
	}
#endif
					
					oplus[it->x][it->y] = idx;
					set<todoElem>::iterator itpom2  = it;
					++itpom2;
					gen_oplus_from_set_todo_with_depMap_with_assocCheckMap(itpom2, n, set_todo, ordersmall, oplus, perm_left, perm_right, setWPES, setISOS, dependencyMap, assocCheckMapCopy);
					oplus[it->x][it->y] = -1;
				}
			}
		}
	}
}

// old method without checking associativity
void gen_oplus_from_set_todo_with_depMap(set<todoElem>::iterator it, signed char n, set<todoElem> &set_todo, int ordersmall[MAX][MAX], signed char oplus[MAX][MAX], signed char perm_left[MAX], signed char perm_right[MAX], set<struct wprepea> &setWPES, set<struct wprepea> &setISOS, const map<todoElem, set<todoElem> > &dependencyMap) {

	if (it == set_todo.end()) {
		if (no_of_iterations == ITERCOUNT) {
			time_t tim = time(NULL);
			cerr << "[" << tim << "]" << endl;
			no_of_iterations = 0;
			cerr << "\tState: ";
			for (set<todoElem>::iterator itp = set_todo.begin(); itp != set_todo.end(); ++itp)
				cerr << setw(2) << (int)oplus[itp->x][itp->y];
			cerr << endl;	
		}
		no_of_iterations++;

		if (is_weak_pre_pea(n, oplus, perm_right, perm_left)) {

			struct wprepea wp(n, oplus, perm_left, perm_right);
			set<struct wprepea>::iterator it1, it2;

			it1 = setWPES.find(wp);

			it2 = setISOS.find(wp);

			if (it1 == setWPES.end() && it2 == setISOS.end()) {

				setWPES.insert(wp);

				print_op(cout, n, oplus);

				cout << "Left orthosupplement:"<<endl;
				for (int i = 0; i < n; i++)
					cout << setw(3) << (int)perm_left[i];
				cout << endl;

				cout << "Right orthosupplement:"<<endl;
				for (int i = 0; i < n; i++)
					cout << setw(3) << (int)perm_right[i];
				cout << endl;
				
#ifdef DEBUG
				cout << "Storing into WPES set." << endl;
#endif

				if (condition(n, oplus))
					cout << "Pre pseudo effect algebra." << endl;
				cout << endl;	

				update_ISOS(ordersmall, wp, setWPES, setISOS);
			}


		}
	}	
	else {

		signed char atleast = max(it->x, it->y), 
		    atmost = n - 1;
/*
		for (set<todoElem>::iterator itpom = set_todo.begin(); itpom != it; itpom++) {
			if (ordersmall[itpom->x - 1]
				      [it->x - 1] 
				&& 
			    ordersmall[itpom->y -1]
				      [it->y - 1]
			   ) {
				atleast = max(atleast, oplus[itpom->x][itpom->y]);

			}
*/

#if 0
			if (ordersmall[it->x - 1]
				      [itpom->x - 1] 
				&& 
			    ordersmall[it->y -1]
				      [itpom->y - 1]
			   ) {
				atmost = min(atmost, oplus[itpom->x][itpom->y]);

			}
#endif


	//	}

//#ifdef DEBUG
		cout /*<< "Level: " << level << endl*/ << "At least: " << atleast << endl << "At most: " << atmost << endl;
//#endif

		for (int idx = atleast ; idx <= atmost; idx++) {
			if (	idx != it->x && idx != it->y &&
 				(atleast == n - 1 || idx == n - 1 || ordersmall[atleast - 1][idx - 1]) 
				&& (idx == n - 1 || ordersmall[it->x - 1][idx - 1]) 
				&& (idx == n - 1 || ordersmall[it->y - 1][idx - 1])
				&& (idx == n - 1 || isFromUpperBound(dependencyMap, *it, idx, n, oplus, ordersmall))
			) {
				oplus[it->x][it->y] = idx;
				set<todoElem>::iterator itpom2  = it;
				++itpom2;
				gen_oplus_from_set_todo_with_depMap(itpom2, n, set_todo, ordersmall, oplus, perm_left, perm_right, setWPES, setISOS, dependencyMap);
				oplus[it->x][it->y] = -1;
			}
		}
	}
}

void print_set_todo(set<todoElem> &set_todo) {
	cerr << "n: " << set_todo.size() << endl;
	for (set<todoElem>::iterator it = set_todo.begin(); it != set_todo.end(); ++it) 
		cerr << "(" << (int)it->x << "," << (int)it->y << "), ";
	
	cerr << endl;
}

void gen_all_permutations_set_depMap(int level, signed char n, signed char perm[MAX], int ordersmall[MAX][MAX], set<struct wprepea> &setWPES, set<struct wprepea> &setISOS, set<setpermElem> &setPERM) {
	int tmp;

	if (level == n-1) {
		signed char right_ortho[MAX], oplus[MAX][MAX];

		compute_inverse_perm(n, perm, right_ortho);

		set<todoElem> test2;

//#ifdef DEBUG
#if 0
		cerr << "Current permutation: ";
		for (int i = 1; i < n-1;i++)
			cerr << setw(3) << (int)perm[i];
		cerr << endl;
#endif
/*
		cout << "-1" << endl;
		for (int i = 1; i < n-1;i++)
			cout << setw(3) << right_ortho[i];
		cout << endl;
*/

//		if (
		
//		for (int idx = 0; idx < order6count; idx++) {

			if (test_order_ortho(n, ordersmall, perm, right_ortho)) {


				if (
				 init_set_todo(n, test2, ordersmall, perm, right_ortho) &&
					setPERM.find(setpermElem(n,perm)) == setPERM.end()
				) {

					cerr << "Tested permutation: ";
					for (int i = 1; i < n-1;i++)
						cerr << setw(3) << (int)perm[i];
					cerr << endl;

						set_initial_plus(n, oplus);
			#ifdef DEBUG
						cerr << "set_initial_plus" << endl;
						print_op(cerr, n, oplus);
			#endif

						init_ones(n, oplus, perm, right_ortho);

			#ifdef DEBUG
						cerr << "init_ones" << endl;
						print_op(cerr, n, oplus);
			#endif

					//init_set_todo(n, test2, ordersmall, perm, right_ortho);
	#if 1
					cerr << "Todo elements: ";
					print_set_todo(test2);
	#endif
					map<todoElem, set<todoElem> > depMap;

					init_dependency_map(test2, ordersmall, depMap);
					cerr << "Dependency map:" << endl;
					display_dependency_map(cerr, depMap);
					cerr << "---" << endl;
					reduce_dependency_map(depMap);
					cerr << "Reduced dependency map:" << endl;
					display_dependency_map(cerr, depMap);
					cerr << "---" << endl;
					


					signed char perm_iso[MAX];
					for (int i = 0; i < n; i++) perm_iso[i] = i;
					update_set_perm(1, ordersmall, setPERM, n, perm, perm_iso);

#ifndef PLAN
					time_t tim = time(NULL);
					cerr << "[" << tim << "]" << endl;
					no_of_iterations = 0UL;
					gen_oplus_from_set_todo_with_depMap(test2.begin(), n, test2, ordersmall, oplus, perm, right_ortho, setWPES, setISOS, depMap);
					tim = time(NULL);
					cerr << "[" << tim << "]" << endl;
					cerr << "\tRemaining count: " << no_of_iterations << endl;
					
#endif
				}
			}
//		}


	}
	else {
		for (int idx=level; idx < n-1; idx++) {
			if (idx != level) {
				tmp = perm[idx];
				perm[idx] = perm[level];
				perm[level] = tmp;

			}
			gen_all_permutations_set_depMap(level+1, n, perm, ordersmall, setWPES, setISOS, setPERM);

			if (idx != level) {
				tmp = perm[idx];
				perm[idx] = perm[level];
				perm[level] = tmp;
			}
		}
	}	


}

void gen_all_permutations_set_depMap_AssocCheckMap(int level, signed char n, signed char perm[MAX], int ordersmall[MAX][MAX], set<struct wprepea> &setWPES, set<struct wprepea> &setISOS, set<setpermElem> &setPERM) {
	int tmp;

	if (level == n-1) {
		signed char right_ortho[MAX], oplus[MAX][MAX];

		compute_inverse_perm(n, perm, right_ortho);

		set<todoElem> test2;

//#ifdef DEBUG
#if 0
		cerr << "Current permutation: ";
		for (int i = 1; i < n-1;i++)
			cerr << setw(3) << (int)perm[i];
		cerr << endl;
#endif
/*
		cout << "-1" << endl;
		for (int i = 1; i < n-1;i++)
			cout << setw(3) << right_ortho[i];
		cout << endl;
*/

//		if (
		
//		for (int idx = 0; idx < order6count; idx++) {

			if (test_order_ortho(n, ordersmall, perm, right_ortho)) {


				if (
				 init_set_todo(n, test2, ordersmall, perm, right_ortho) &&
					setPERM.find(setpermElem(n,perm)) == setPERM.end()
				) {

					cerr << "Tested permutation: ";
					for (int i = 1; i < n-1;i++)
						cerr << setw(3) << (int)perm[i];
					cerr << endl;

						set_initial_plus(n, oplus);
			#ifdef DEBUG
						cerr << "set_initial_plus" << endl;
						print_op(cerr, n, oplus);
			#endif

						init_ones(n, oplus, perm, right_ortho);

			#ifdef DEBUG
						cerr << "init_ones" << endl;
						print_op(cerr, n, oplus);
			#endif

					//init_set_todo(n, test2, ordersmall, perm, right_ortho);
	#if 1
					cerr << "Todo elements: ";
					print_set_todo(test2);
	#endif
					map<todoElem, set<todoElem> > depMap;

					init_dependency_map(test2, ordersmall, depMap);
					cerr << "Dependency map:" << endl;
					display_dependency_map(cerr, depMap);
					cerr << "---" << endl;
					reduce_dependency_map(depMap);
					cerr << "Reduced dependency map:" << endl;
					display_dependency_map(cerr, depMap);
					cerr << "---" << endl;
					


					signed char perm_iso[MAX];
					for (int i = 0; i < n; i++) perm_iso[i] = i;
					update_set_perm(1, ordersmall, setPERM, n, perm, perm_iso);

#ifndef PLAN
					time_t tim = time(NULL);
					cerr << "[" << tim << "]" << endl;
					no_of_iterations = 0UL;

					map< pair<signed char, signed char>, set<signed char> > assocCheckMap;
					gen_oplus_from_set_todo_with_depMap_with_assocCheckMap(test2.begin(), n, test2, ordersmall, oplus, perm, right_ortho, setWPES, setISOS, depMap, assocCheckMap);

					tim = time(NULL);
					cerr << "[" << tim << "]" << endl;
					cerr << "\tRemaining count: " << no_of_iterations << endl;
					
#endif
				}
			}
//		}


	}
	else {
		for (int idx=level; idx < n-1; idx++) {
			if (idx != level) {
				tmp = perm[idx];
				perm[idx] = perm[level];
				perm[level] = tmp;

			}
			gen_all_permutations_set_depMap_AssocCheckMap(level+1, n, perm, ordersmall, setWPES, setISOS, setPERM);

			if (idx != level) {
				tmp = perm[idx];
				perm[idx] = perm[level];
				perm[level] = tmp;
			}
		}
	}	


}

int main(int argc, char* argv[]) {

	int n = N;

	signed char perm[MAX];
	perm[0] = n - 1; perm[n - 1] = 0;

	if (argc < 2) {
		cout << "Usage: " << argv[0] << " <order_specification>" << endl;
		cout << "\t<order_specification>:"<< endl;
		cout << "\t\t- = test orders from " << VFROM + 1 << " to " << VTO << endl;
		cout << "\t\tord_num- = test orders from ord_num to " << VTO << endl;
		cout << "\t\t-ord_num = test orders from " << VFROM + 1 << "to ord_num" << endl;
		cout << "\t\tord_num = test order ord_num" << endl;
		cout << "\t\tord_num1-ord_num2 = test orders from ord_num1 to ord_num2" << endl;
		return 1;
	}

	int from, to;

	if (parse_order_specification(argv[1], VFROM, VTO, from, to)) {

#ifdef DEBUG
		cout << "from: " << from << ", to: " << to << endl;
#endif

		from--;
		
		if (from < VFROM || to > VTO || from > to) {
			cout << "Allowed range is: " << VFROM + 1 << "-" << VTO << "." << endl;
			return 1;
		}

		for (int idx = (from-order_offset); idx < (to-order_offset); idx++) {
			set<struct wprepea> setWPES, setISOS;
			set<setpermElem> setPERM;
			
			for (int i = 1; i < n - 1; i++) perm[i] = i;

			cout << "Order n. " << order_offset + idx + 1 << endl;
			cerr << "Order n. " << order_offset + idx + 1 << endl;

	#ifdef PLAN
			for (int a = 0; a < n - 2; a++) {
				for (int b = 0; b < n - 2; b++)
					cerr << setw(2) << ORDERSARRAY[idx][a][b];
				cerr << endl;
			}
	#endif


			gen_all_permutations_set_depMap_AssocCheckMap(1, n, perm, ORDERSARRAY[idx], setWPES, setISOS, setPERM);
		}
	}

	return 0;
}
