#include <iostream>
#include "consts.h"
#include "wprepea_parseutils.h"


using namespace std;

bool parse_permutation(const char *permspec, const int nmax, signed char *perm) {
	int len = 0;
	const char *pom = permspec;

	while (*pom && len <= nmax - 2) {
		if (*pom>='0' && *pom<='9') {
			signed char num = 0;
			while (*pom>= '0' && *pom <= '9') {
				num = 10*num + (*pom - '0');
				pom++;
			}

			if (num >= 1 && num <= nmax - 2) {
				perm[1 + len] = num;
				len++;
			}
			else {
				cerr << "Elements in permutation are allowed in range: " << 1 << "-" << nmax - 2 <<"." << endl;
				return false;
			}
		}
		else {
			cerr << "Unknown character: '" << *pom << "'." << endl;
				cerr << permspec << endl;
				for (int i = 0; i < (pom - permspec); i++)
					cerr << " ";
				cerr << "^" << endl;

			return false;
		}

		if (len <= nmax - 3) {
			if (*pom != ',') {
				if (!*pom) {
					//no more characters
					cerr << "Permutation specification too short." << endl;
					return false;
				}

				cerr << "Unknown character, expecting comma." << endl;
	
				return false;
			}
			pom++;
		}

/*
		if (len > N - 2) {
			cerr << "Permutation specification too long." << endl;
			return false;
		}
*/
	}
	
	if (len < nmax - 2) {
		cerr << "Permutation specification too short." << endl;
		return false;
	}

/*
	if (*pom) {
		cerr << "[Warning] Trailing character in permutation specification. '" << (int) *pom << "'" << endl;
	}
*/
	return true;
	
}

bool is_permutation(int n, signed char *perm) {
	bool in_set[MAX];

	for (int i = 0; i < n; i++)
		in_set[i] = false;

	for (int i = 0; i < n; i++)
		if (in_set[perm[i]] == false)
			in_set[perm[i]] = true;
		else {
			cerr << "Not a permutation. Element " << (int) perm[i] << " is already in." << endl;
			return false;
		}
	return true;
}

bool parse_state(const char *statespec, const int maxlength, const int nmax, int &state_len, int *state, bool* was_used) {
	const char *pom  = statespec;
	int state_length = 0;
	int read_length = 0;


	if (*pom == '-') {
		// from beginning
		state_len = 0;
		return true;
	}


	if (*pom >='0' && *pom <='9') {
		while (*pom >='0' && *pom <='9') {
			state_length = 10 * state_length + (*pom - '0');	
			pom++;
		}
		if (*pom != ':' || !*pom) {
			cerr << "Expecting :" << endl;
			return false;
		}
		pom++;

		if (state_length > maxlength) {
			cerr << "Error: state_length should be shorter than " << maxlength << endl;
			cerr << "You need to change constant maxdepth and recompile." << endl;
			return false;
		}

		while (*pom && read_length < state_length) {

			if (*pom >= '0' && *pom <='9') {
				int num = 0;
				while (*pom >='0' && *pom <= '9') {
					num = 10*num + (*pom - '0');
					pom++;
				}
				if (num < 1 || num > nmax - 1) {
					cerr << "Allowed number is between 1 - " << nmax - 1 << endl;
					return false;
				}

				state[read_length] = num;
				was_used[read_length] = false;
				read_length++;
					
			}
			else {
				cerr << "Expecting number" << endl;
				return false;
			}

			if (read_length < state_length && *pom != ',') {
				cerr << "Expecting comma" << endl;
				return false;
			}
			
			pom++;

		}
	}
	else {
		cerr << "Expecting the length of the state." << endl;
		return false;
	}

	if (read_length != state_length) {
		cerr << "Only " << read_length << " elements are present. Expecting " << state_length << " elements" << endl;
		return false;
	}
	

	state_len = state_length;
	return true;
}

bool parse_order_specification(char *spec, const int nfrom, const int nto, int &from, int &to) {
	char *pom = spec;

	if (*pom == '-') {
		// no starting range
		from = nfrom + 1;
		pom++;
		if (!*pom) {
			// no end range
			to = nto;
		} else {
			if (*pom >= '0' && *pom <= '9') {
				// read number to
				int num = 0;
				while (*pom >= '0' && *pom <= '9') {
					num = 10 * num + (*pom - '0');
					pom++;
				}
				to = num;
			}
			else {
				cerr << "Unknown character '" << *pom << "'." << endl;
				return false;
			}
		}
	
	} else {
		if (*pom >= '0' && *pom <= '9') {
			// read number from
			int num = 0;
			while (*pom >= '0' && *pom <= '9') {
				num = 10 * num + (*pom - '0');
				pom++;
			}
			from = num;
			
		}
		else {
			cerr << "Unknown character '" << *pom << "'." << endl;
			return false;
		}
		if (!*pom) {
			//single number
			to = from;
		} else {
			if (*pom == '-') {
				pom++;
				if (!*pom) {
					//range from-DTVO
					to = nto;
				} else {
					if (*pom >='0' && *pom <= '9') {
						// read number to
						int num = 0;
						while (*pom >= '0' && *pom <= '9') {
							num = 10 * num + (*pom - '0');
							pom++;
						}
						to = num;
					}
					else {
						cerr << "Unknown character '" << *pom << "'." << endl;
						return false;
					}
				}
			}
			else {
				cerr << "Unknown character '" << *pom << "'." << endl;
				return false;
			}

		}
	}


	return true;
}
