#include "wprepea_utils.h"
#include <iomanip>

permutation::permutation() {
	length = 0;
}

permutation::permutation(int length, int perm[MAX]) {
	this->length = length;
	for (int i = 0; i < length; i++)
		this->perm[i] = perm[i]; 
}

permutation::permutation(const struct permutation &perm) {
	length = perm.length;

	for (int i = 0; i < perm.length; i++)
		this->perm[i] = perm.perm[i]; 
}



//using namespace std;

void update_ISOS(int ordersmall[MAX][MAX], const struct wprepea wp, set<struct wprepea> &setWPES, set<struct wprepea> &setISOS) {

	int isoperm[MAX];

	for (int i = 0; i < wp.n ; i++)
		isoperm[i] = i;

	update_ISOS_recursive(1, ordersmall, wp, isoperm, setWPES, setISOS);
}


void update_ISOS_recursive(int level, int ordersmall[MAX][MAX], const struct wprepea wp, int perm[MAX], set<struct wprepea> &setWPES, set<struct wprepea> &setISOS) {
	
	if (level == wp.n - 2) {
		struct wprepea wp2;

		signed char applied_perm[MAX];

		applied_perm[0] = wp.n - 1;	
		applied_perm[wp.n - 1] = 0;	

		for (int i = 1; i < wp.n - 1; i++) 
			applied_perm[perm[i]]  = perm[wp.left_ortho[i]];	

		if (is_admissible(wp.n, ordersmall, applied_perm)) {

			isoimage(wp, perm, wp2);


	#ifdef DEBUG
			cout << "\tISO:";
			for (int i = 0; i < wp.n; i++) cout << perm[i];
			cout  << endl;
			print_op(wp2.n, wp2.oplus);
			cout << "Left orthosupplement:"<<endl;
			for (int i = 0; i < wp2.n; i++)
				cout << setw(3) << (int)wp2.left_ortho[i];
			cout << endl;

			cout << "Right orthosupplement:"<<endl;
			for (int i = 0; i < wp2.n; i++)
				cout << setw(3) << (int)wp2.right_ortho[i];
			cout << endl;
	#endif



			set<struct wprepea>::iterator it1, it2;

	#ifdef DEBUG
			cout << "Size WPES: " << setWPES.size() << endl;
	#endif

			it1 = setWPES.find(wp2);

	#ifdef DEBUG
			if (it1 != setWPES.end())
				cout << "Found in WPES." << endl;
			else {
				cout << "Not found in WPES." << endl;

				cout << "Set content: " << endl;
				for (it1 = setWPES.begin(); it1 != setWPES.end(); ++it1) {

					print_op(it1->n, it1->oplus);
					cout << "Left orthosupplement:"<<endl;
					for (int i = 0; i < it1->n; i++)
						cout << setw(3) << it1->left_ortho[i];
					cout << endl;

					cout << "Right orthosupplement:"<<endl;
					for (int i = 0; i < it1->n; i++)
						cout << setw(3) << it1->right_ortho[i];
					cout << endl;

					if (wp2 == *it1)
						cout << "Zhoda" << endl;
					else
						cout << "Nezhoda" << endl;
				
					
				}
			}
	#endif

			it2 = setISOS.find(wp2);

			if (it1 == setWPES.end() && it2 == setISOS.end()) {

	#ifdef DEBUG
				cout << "Size before: " << setISOS.size() << endl;
	#endif
				setISOS.insert(wp2);
	#ifdef DEBUG
				cout << "Size after: " << setISOS.size() << endl;
	#endif
			}

	#ifdef DEBUG
			else 
				cout << "Already in set." << endl;
		
	#endif
		}
	}
	else {
		for (int k = level;  k < wp.n - 1; k++) {
			if (level != k) {
				int temp = perm[level];
				perm[level] = perm[k];
				perm[k] = temp;
			}
			update_ISOS_recursive(level + 1, ordersmall, wp, perm, setWPES, setISOS);
			if (level != k) {
				int temp = perm[level];
				perm[level] = perm[k];
				perm[k] = temp;
			}
		
		}
	}

}

void update_ISOS_with_map(int ordersmall[MAX][MAX], const struct wprepea wp, set<struct wprepea> &setWPES, set<struct wprepea> &setISOS, map<wprepea, pair<wprepea,permutation> > &isomap) {

	int isoperm[MAX];

	for (int i = 0; i < wp.n ; i++)
		isoperm[i] = i;

	update_ISOS_with_map_recursive(1, ordersmall, wp, isoperm, setWPES, setISOS, isomap);
}


void update_ISOS_with_map_recursive(int level, int ordersmall[MAX][MAX], const struct wprepea wp, int perm[MAX], set<struct wprepea> &setWPES, set<struct wprepea> &setISOS, map<wprepea, pair<wprepea, permutation> > &isomap) {
	
	if (level == wp.n - 2) {
		struct wprepea wp2;

		signed char applied_perm[MAX];

		applied_perm[0] = wp.n - 1;	
		applied_perm[wp.n - 1] = 0;	

		for (int i = 1; i < wp.n - 1; i++) 
			applied_perm[perm[i]]  = perm[wp.left_ortho[i]];	

		if (is_admissible(wp.n, ordersmall, applied_perm)) {

			isoimage(wp, perm, wp2);


	#ifdef DEBUG
			cout << "\tISO:";
			for (int i = 0; i < wp.n; i++) cout << perm[i];
			cout  << endl;
			print_op(wp2.n, wp2.oplus);
			cout << "Left orthosupplement:"<<endl;
			for (int i = 0; i < wp2.n; i++)
				cout << setw(3) << (int)wp2.left_ortho[i];
			cout << endl;

			cout << "Right orthosupplement:"<<endl;
			for (int i = 0; i < wp2.n; i++)
				cout << setw(3) << (int)wp2.right_ortho[i];
			cout << endl;
	#endif



			set<struct wprepea>::iterator it1, it2;

	#ifdef DEBUG
			cout << "Size WPES: " << setWPES.size() << endl;
	#endif

			it1 = setWPES.find(wp2);

	#ifdef DEBUG
			if (it1 != setWPES.end())
				cout << "Found in WPES." << endl;
			else {
				cout << "Not found in WPES." << endl;

				cout << "Set content: " << endl;
				for (it1 = setWPES.begin(); it1 != setWPES.end(); ++it1) {

					print_op(it1->n, it1->oplus);
					cout << "Left orthosupplement:"<<endl;
					for (int i = 0; i < it1->n; i++)
						cout << setw(3) << it1->left_ortho[i];
					cout << endl;

					cout << "Right orthosupplement:"<<endl;
					for (int i = 0; i < it1->n; i++)
						cout << setw(3) << it1->right_ortho[i];
					cout << endl;

					if (wp2 == *it1)
						cout << "Zhoda" << endl;
					else
						cout << "Nezhoda" << endl;
				
					
				}
			}
	#endif

			it2 = setISOS.find(wp2);

			if (it1 == setWPES.end() && it2 == setISOS.end()) {

	#ifdef DEBUG
				cout << "Size before: " << setISOS.size() << endl;
	#endif
				setISOS.insert(wp2);


				isomap.insert( pair<wprepea, pair<wprepea, permutation> >(wp2, pair<wprepea, permutation>(wp, permutation(wp.n, perm))) );
	#ifdef DEBUG
				cout << "Size after: " << setISOS.size() << endl;
	#endif
			}

	#ifdef DEBUG
			else 
				cout << "Already in set." << endl;
		
	#endif
		}
	}
	else {
		for (int k = level;  k < wp.n - 1; k++) {
			if (level != k) {
				int temp = perm[level];
				perm[level] = perm[k];
				perm[k] = temp;
			}
			update_ISOS_with_map_recursive(level + 1, ordersmall, wp, perm, setWPES, setISOS, isomap);
			if (level != k) {
				int temp = perm[level];
				perm[level] = perm[k];
				perm[k] = temp;
			}
		
		}
	}

}


bool is_admissible(signed char n, int ordersmall[MAX][MAX], signed char perm[MAX]) {

	signed char right_ortho[MAX];
	set<todoElem> todo;

	compute_inverse_perm(n, perm, right_ortho);

	
	return init_set_todo(n, todo, ordersmall, perm, right_ortho);

#if 0
	if (!init_set_todo(n, todo, ordersmall, perm, right_ortho))
		return false;
	else {
		for (int i=1; i < n - 1; i++)
		for (int j=1; j < n  -1; j++)
			if (i != j) {
				if (
					(ordersmall[i - 1][j - 1] && !ordersmall[perm[i] - 1][perm[j] -1]) ||
					(!ordersmall[i - 1][j - 1] && ordersmall[perm[i] - 1][perm[j] - 1])
				)
					return false;
		}
		return true; 
	}
#endif 
}

void isoimage(const struct wprepea wp1, int isomorphism[MAX], struct wprepea &wp2) {
	
	wp2.n = wp1.n;

	set_initial_plus(wp1.n, wp2.oplus);

	for (int i = 1; i < wp1.n - 1 ; i++)
	for (int j = 1; j < wp1.n - 1 ; j++)
	{
		if (wp1.oplus[i][j] == -1) 
			wp2.oplus[isomorphism[i]][isomorphism[j]] = -1;
		else		
			wp2.oplus[isomorphism[i]][isomorphism[j]] = isomorphism[wp1.oplus[i][j]];
	}

	wp2.left_ortho[0] = wp2.right_ortho[0] = wp1.n - 1;	
	wp2.left_ortho[wp1.n - 1] = wp2.right_ortho[wp1.n - 1] = 0;	

	for (int i = 1; i < wp1.n - 1; i++) {
		wp2.left_ortho[isomorphism[i]]  = isomorphism[wp1.left_ortho[i]];	
		wp2.right_ortho[isomorphism[i]] = isomorphism[wp1.right_ortho[i]];	

	}
}

void compute_inverse_perm(signed char n, signed char perm[MAX], signed char perm_inv[MAX]) {

	for (int idx = 0; idx < n ; idx++)
		perm_inv[perm[idx]] = idx;
}

bool init_set_todo(signed char n, set<todoElem> &set_todo, int ordersmall[MAX][MAX], signed char left_ortho[MAX], signed char right_ortho[MAX]) {

	set_todo.clear();

	for (int i = 0; i < n - 2; i++)
	for (int j = i+1; j < n - 2; j++) {
		if (ordersmall[i][j]) {
			//insert (i, rightortho[j]), (leftortho[j], i) into todo_list

			set_todo.insert(todoElem(i + 1, right_ortho[j + 1]));
			set_todo.insert(todoElem(left_ortho[j + 1], i + 1));
		}
	}

	for (int i = 0; i < n - 2; i++)
	for (int j = 0; j < n - 2; j++) {
		if (i != j && !ordersmall[i][j]) {

			if (set_todo.find(todoElem(i + 1, right_ortho[j + 1])) != set_todo.end())
				return false;

			if (set_todo.find(todoElem(left_ortho[j + 1], i + 1)) != set_todo.end())
				return false;

			if (set_todo.find(todoElem(left_ortho[left_ortho[i + 1]], left_ortho[j + 1])) != set_todo.end())
				return false;

			if (set_todo.find(todoElem(right_ortho[j + 1], right_ortho[right_ortho[i + 1]])) != set_todo.end())
				return false;
			
		}
	}


	return true;
}


void set_initial_plus(signed char n, signed char op_plus[MAX][MAX]) {

	for (int i = 0; i < n; i++)
		for (int j = 0; j < n; j++) {
			set_oplus(op_plus, i, j, -1);
		}

	for (int i = 0; i < n; i++) {
		set_oplus(op_plus, i, 0, i);
		set_oplus(op_plus, 0, i, i);
	}
}

bool operator<(const struct todoElem &te1, const struct todoElem &te2) {
	if ((te1.x + te1.y < te2.x + te2.y) || ((te1.x + te1.y == te2.x + te2.y) && (te1.x < te2.x || (te1.x == te2.x && te1.y < te2.y))))
		return true;

	return false;
}

void init_dependency_map(const set<todoElem> set_todo, int ordersmall[MAX][MAX], map<todoElem, set<todoElem> > &dependencyMap) {
	

	for (set<todoElem>::iterator it = set_todo.begin(); it != set_todo.end(); ++it) {
		
		set<todoElem> tempSet;
		bool isEmpty = true;

		for (set<todoElem>::iterator it_bot = set_todo.begin(); it_bot != it; ++it_bot) {
			if (ordersmall[it_bot->x - 1][it->x - 1] && ordersmall[it_bot->y - 1][it->y - 1]) {
				tempSet.insert(*it_bot);
				isEmpty = false;
			}
		}

		if (!isEmpty) 
			dependencyMap.insert(pair<todoElem, set<todoElem> >(*it, tempSet));	
	}
}

void reduce_dependency_map(map<todoElem, set<todoElem> > &dependencyMap) {
	map<todoElem, set<todoElem> >::iterator it;

	for (map<todoElem, set<todoElem> >::iterator itmap = dependencyMap.begin(); itmap != dependencyMap.end(); ++itmap) {
		for (set<todoElem>::iterator itset = (itmap->second).begin(); itset != (itmap->second).end(); ++itset) {
			if ((it = dependencyMap.find(*itset)) != dependencyMap.end()) {
				for (set<todoElem>::iterator itremset = (it->second).begin(); itremset != (it->second).end(); ++itremset)
					(itmap->second).erase(*itremset);
			}
						
		}
	}
}



void display_dependency_map(ostream &os, const map<todoElem, set<todoElem> > &dependencyMap) {
	
	for (map<todoElem, set<todoElem> >::const_iterator it = dependencyMap.begin(); it != dependencyMap.end(); ++it) {
		os << "(" << (int)(it->first).x << "," << (int)(it->first).y << "): ";

		for (set<todoElem>::const_iterator setit = (it->second).begin(); setit != (it->second).end(); ++setit) {
			os << "(" << (int)setit->x << "," << (int)setit->y << "), ";
		}
		os << endl;
	}
}

bool operator<(const struct setpermElem &spe1, const struct setpermElem &spe2) {
	if (spe1.n < spe2.n)
		return true;
	if (spe1.n > spe2.n)
		return false;

	for (int i = 0; i < spe1.n ; i++)
		if (spe1.perm[i] != spe2.perm[i]) {
			if (spe1.perm[i] < spe2.perm[i]) 
				return true;
			else 
				return false;
		}

	return false;
}


bool preserve_order(int n, int ordersmall[MAX][MAX], signed char perm[MAX]) {
	for (int ii=1; ii < n - 1; ii++)
	for (int jj=1; jj < n  -1; jj++)
		if (ii != jj) {
			if (
				(ordersmall[ii - 1][jj - 1] && !ordersmall[perm[ii] - 1][perm[jj] -1]) ||
				(!ordersmall[ii - 1][jj - 1] && ordersmall[perm[ii] - 1][perm[jj] - 1])
			) {
				return false;
			}
		}
	return true;
}

void update_set_perm(int level, int ordersmall[MAX][MAX], set<setpermElem> &setPERM, int n, signed char perm[MAX], signed char perm_iso[MAX], const bool is_plan) {

	if (level== n - 1) {
		
		signed char applied_perm[MAX];

		applied_perm[0] = n - 1;	
		applied_perm[n - 1] = 0;	

		for (int i = 1; i < n - 1; i++) 
			applied_perm[perm_iso[i]]  = perm_iso[perm[i]];	


		if (is_admissible(n, ordersmall, applied_perm)) {

			if (preserve_order(n, ordersmall, perm_iso)) {

#ifdef DEBUG
		cerr << "Perm ISO: ";
		for (int i = 1; i < n - 1; i++)
			cerr << setw(2) << (int) perm[i] ;
		cerr <<endl;
		cerr << "Applied Perm: ";
		for (int i = 1; i < n - 1; i++)
			cerr << setw(2) << (int) applied_perm[i] ;
		cerr <<endl;
#endif
#if 0
				cerr << "Is in set?: " ;
				for (int i = 1; i < n - 1; i++) {
					cerr << setw(3) << (int) applied_perm[i];
					
				}
				cerr << endl;
#endif

#if 0
		for (int i=1; i < n - 1; i++)
		for (int j=1; j < n  -1; j++)
			if (i != j) {
				if (
					(ordersmall[i - 1][j - 1] && !ordersmall[perm[i] - 1][perm[j] -1]) ||
					(!ordersmall[i - 1][j - 1] && ordersmall[perm[i] - 1][perm[j] - 1])
				)
					return false;
		}
		return true; 
	
#endif 
		


				if (setPERM.find(setpermElem(n, applied_perm)) == setPERM.end()) {

					if (is_plan) {
						cerr << "\t Isomorphic perm: " ;
						for (int i = 1; i < n - 1; i++) {
							cerr << setw(3) << (int) applied_perm[i];
							
						}
						cerr << " ("; 
						for (int i = 1; i < n - 1; i++) {
							cerr << setw(2) << (int) perm_iso[i];
							
						}
						cerr << ")";

						cerr << endl;
					}

					setPERM.insert(setpermElem(n, applied_perm));
				}
	//			cerr << "setPERM size: " << setPERM.size() << endl;
			}
		}

	}
	else {
		for (int i = level; i < n - 1; i++) {
			if (i != level) {
				signed char tmp = perm_iso[i];
				perm_iso[i] = perm_iso[level];
				perm_iso[level] = tmp;
			}
			update_set_perm(level + 1, ordersmall, setPERM, n, perm, perm_iso);
			if (i != level) {
				signed char tmp = perm_iso[i];
				perm_iso[i] = perm_iso[level];
				perm_iso[level] = tmp;
			}
		}
	}
}

bool test_order_ortho(signed char n, int ordersmall[MAX][MAX], signed char left_ortho[MAX], signed char right_ortho[MAX]) {
	for (int i = 1; i < n - 2; i++)
		for (int j = i + 1; j < n - 1; j++) 
			if (ordersmall[i - 1][j - 1] && (!ordersmall[left_ortho[j] - 1][left_ortho[i] - 1] || !ordersmall[right_ortho[j] - 1][right_ortho[i] - 1]))
				return false;
			
		
	return true;
}

bool isFromUpperBound(const map<todoElem, set<todoElem> > &dependencyMap, const todoElem &elem, int &idx, int n, signed char oplus[MAX][MAX], int ordersmall[MAX][MAX]) {
	map<todoElem, set<todoElem> >::const_iterator it =  dependencyMap.find(elem);

	if (it == dependencyMap.end()) {
		// no dependencies	
		return true;
	}
	else {
		// idx != n - 1		

		for (set<todoElem>::const_iterator its = (it->second).begin(); its != (it->second).end(); ++its) {
			int val = oplus[its->x][its->y];
			if (val == n - 1) {
				// set idx = n - 1 and return
				idx = n - 1;
				return true;
			}
			if (!ordersmall[val - 1][idx - 1])
				return false;
		}
	}
	return true;
}

bool testAssocCheckMap(map< pair<signed char, signed char>, set<signed char> > &assocCheckMap, const todoElem &elem, int &idx, int n, signed char oplus[MAX][MAX], signed char perm_left[MAX], signed char perm_right[MAX], int ordersmall[MAX][MAX]) {

	map< pair<signed char, signed char>, set<signed char> >::const_iterator itAssocMap = assocCheckMap.find(pair<signed char, signed char>(elem.x, elem.y));

	if (itAssocMap != assocCheckMap.end()) {
		for (set<signed char>::const_iterator itset = (itAssocMap->second).begin(); itset != (itAssocMap->second).end(); itset++) {
			if (idx != 0 && *itset != n - 1 && !ordersmall[idx - 1][*itset - 1])
				return false;
		}
	}

	set< pair< pair<signed char, signed char>, signed char> > setToAdd;

	signed char new_pair_x = perm_left[idx], new_pair_y = elem.x;
	if (new_pair_x == elem.x && new_pair_y == elem.y) {
		if (!ordersmall[idx - 1][perm_left[new_pair_y] - 1]) 
			return false;
	}
	else {
		if (is_defined(oplus, new_pair_x, new_pair_y)) {
			signed char value = get_oplus(oplus, new_pair_x, new_pair_y);
			if (value != 0 && perm_left[elem.y] != n - 1 && !ordersmall[value - 1][perm_left[elem.y] - 1]) 
				return false;
		}
		else {
			setToAdd.insert(pair< pair<signed char, signed char>, signed char>(pair<signed char, signed char>(perm_left[idx], elem.x), perm_left[elem.y]));
		}
	}

	new_pair_x = elem.y; new_pair_y = perm_right[idx];

	if (new_pair_x == elem.x && new_pair_y == elem.y) {
		if (!ordersmall[idx - 1][perm_right[new_pair_x] - 1]) {
			return false;
		}
	}
	else {
		if (is_defined(oplus, new_pair_x, new_pair_y)) {
			signed char value = get_oplus(oplus, new_pair_x, new_pair_y);
			if (value != 0 && perm_right[elem.x] != n - 1 && !ordersmall[value - 1][perm_right[elem.x] - 1]) 
				return false;
		}
		else {
			setToAdd.insert(pair< pair<signed char, signed char>, signed char>(pair<signed char, signed char>(elem.y, perm_right[idx]), perm_right[elem.x]));
		}
	}

	
	// modify assocCheckMap	
	map< pair<signed char, signed char>, set<signed char> > assocCheckMapCopy = assocCheckMap;
	bool areModificationsValid = true;

	for (set< pair< pair<signed char, signed char>, signed char> >::const_iterator itsettoadd = setToAdd.begin(); itsettoadd != setToAdd.end() && areModificationsValid; itsettoadd++) {
		map< pair<signed char, signed char>, set<signed char> >::iterator itAssocMap =  assocCheckMapCopy.find(pair<signed char, signed char>(((*itsettoadd).first).first, ((*itsettoadd).first).second));
		
		if (itAssocMap != assocCheckMapCopy.end()) {
			if ((itAssocMap->second).find((*itsettoadd).second) == (itAssocMap->second).end()) {

				bool isNotEmpty = false;

				for (int i = max((itAssocMap->first).first, (itAssocMap->first).second) + 1; i < n - 1 && !isNotEmpty; i++) {
					if (ordersmall[i - 1][(*itsettoadd).second - 1]) {
						bool isLess = true;

						for (set<signed char>::const_iterator itsetpom = (itAssocMap->second).begin(); itsetpom != (itAssocMap->second).end(); itsetpom++)
							if (!ordersmall[i - 1][*itsetpom -1]) {
								isLess = false;
								break;
							}
						
						if (isLess) {
							isNotEmpty = true;
						}
								
					}				
	
				}
				if (isNotEmpty)
					(itAssocMap->second).insert((*itsettoadd).second);
				else
					areModificationsValid = false;
			}
                }
                else {
                        set<signed char> tempSet;
                        tempSet.insert((*itsettoadd).second);

                        assocCheckMapCopy.insert(pair< pair<signed char, signed char>, set<signed char> >(pair<signed char, signed char>(((*itsettoadd).first).first, ((*itsettoadd).first).second), tempSet));
                }
	
	}
	
	if (!areModificationsValid) {
		return false;
	}
	else {
		assocCheckMap = assocCheckMapCopy;
	}

		
	return true;
}

void init_ones(signed char n, signed char oplus[MAX][MAX], signed char perm_left[MAX], signed char perm_right[MAX]) {

	cerr << "Fixed: ";
	for (int idx = 1; idx < n - 1; idx++) {
		oplus[idx][ perm_right[idx] ] = n - 1;
		cerr << "(" << idx << "," << (int)perm_right[idx] << "), ";
	}
	cerr << endl;
}


bool condition(signed char n, signed char op_plus[MAX][MAX]) {
	for (int i = 0; i < n; i++)
	for (int j = 0; j < n; j++)
		if (is_defined(op_plus, i, j)) {
			bool found_c = false;
			int result = get_oplus(op_plus, i, j);

			for (int k = 0; k < n; k++)
				if (get_oplus(op_plus, k, i) == result) {
					found_c = true;
					break;
				}
			if (!found_c)
				return false;

			bool found_d = false;


			for (int k = 0; k < n; k++)
				if (get_oplus(op_plus, j, k) == result) {
					found_d = true;
					break;
				}
			if (!found_d)
				return false;
		}
	return true;
}

bool is_weak_pre_pea(signed char n, signed char oplus[MAX][MAX], signed char compl_right[MAX], signed char compl_left[MAX]) {
	
	if (	test_associativity(n, oplus) /*&& 
		are_left_right_orders_equal(n, oplus, compl_left, compl_right)*/
	)
		return true;

	return false;
}

bool test_associativity(signed char n, signed char op_plus[MAX][MAX]) {
#ifdef DEBUG
	cerr << "Associativity check:" << endl;
#endif

	for (int a = 1; a < n - 1; a++)
		for (int b = 1; b < n - 1; b++)
			for (int c = 1; c < n - 1; c++) {

				if (is_defined(op_plus, a, b) && is_defined(op_plus, get_oplus(op_plus, a, b), c)) {
					if (!is_defined(op_plus, b, c) || !is_defined(op_plus, a, get_oplus(op_plus, b, c))){
#ifdef DEBUG
						cerr << "-Associativity failed for:" << a << ", " << b << ", " << c << "." << endl;
#endif
						return false;
					}
					if (get_oplus(op_plus, get_oplus(op_plus, a, b), c) != 
					    get_oplus(op_plus, a, get_oplus(op_plus, b, c))) {

#ifdef DEBUG
						cerr << "-Associativity failed for:" << a << ", " << b << ", " << c << "." << endl;
#endif
						return false;
					}
				}

				if (is_defined(op_plus, b, c) && is_defined(op_plus, a, get_oplus(op_plus, b, c))) {
					if (!is_defined(op_plus, a, b) || !is_defined(op_plus, get_oplus(op_plus, a, b), c)) {
#ifdef DEBUG
						cerr << "-Associativity failed for:" << a << ", " << b << ", " << c << "." << endl;
#endif
						return false;
					}
					if (get_oplus(op_plus, get_oplus(op_plus, a, b), c) != 
					    get_oplus(op_plus, a, get_oplus(op_plus, b, c))) {
#ifdef DEBUG

						cout << "-Associativity failed for:" << a << ", " << b << ", " << c << "." << endl;
#endif
						return false;
					}
				}
			}
#ifdef DEBUG
	cerr << "- OK." << endl;
#endif
	return true;
}

void print_op(ostream &os, signed char n, const signed char op[MAX][MAX]) {
	for (int i = 0; i < n; i++) {
		for (int j = 0; j < n; j++) {

			if (is_defined(op, i, j))
				os << setw(3) << (int)get_oplus(op, i, j);
			else
				os << setw(3) << ".";
		}
		os << endl;
	}
}

bool are_left_right_orders_equal(signed char n, signed char op_plus[MAX][MAX], signed char left_ortho[MAX], signed char right_ortho[MAX]) {
	for (int elem_a = 0; elem_a < n; elem_a++)
	for (int elem_b = 0; elem_b < n; elem_b++)
		if ((is_defined(op_plus, elem_a, right_ortho[elem_b]) && !is_defined(op_plus, left_ortho[elem_b], elem_a)) ||
			(is_defined(op_plus, left_ortho[elem_b], elem_a) && !is_defined(op_plus, elem_a, right_ortho[elem_b]))
		)
			return false;

	return true;
}
